""" Unit tests for the the 'geospatial.interpolation' module in the 'kadlu' package

    Authors: Oliver Kirsebom
    contact: oliver.kirsebom@dal.ca
    Organization: MERIDIAN-Intitute for Big Data Analytics
    Team: Acoustic data Analytics, Dalhousie University
    Project: packages/kadlu
             Project goal: Tools for underwater soundscape modeling

    License:

"""
import pytest
import os
import numpy as np
import kadlu
import kadlu.geospatial.interpolation as ki
from kadlu.utils import load_data_from_file, center_point, LLtoXY

path_to_assets = os.path.join(os.path.dirname(os.path.dirname(__file__)), "assets")

np.random.seed(1)

gitlab = os.environ.get("GITLAB_CI") is not None


def test_derivative():
    """ Check interpolation of derivates"""
    # with constant value
    v0 = 12.2
    interp = ki.IrregularGridGeospatialInterpolator(value=v0)
    dv = interp(lon=1.1, dlat=1)
    assert dv == 0

    # with irregular grid 
    v0 = [0, 10, 20]
    lat0 = [0, 1, 2]
    interp = ki.IrregularGridGeospatialInterpolator(value=v0, lat=lat0)
    with pytest.raises(NotImplementedError):
        interp(lat=1.5, dlat=2)

    # with regular grid 
    v0 = [0, 10, 20]
    lat0 = [0, 1, 2]
    interp = ki.RegularGridGeospatialInterpolator(value=v0, lat=lat0)
    with pytest.raises(NotImplementedError):
        interp(lat=1.5, dlat=2)

    # with regular grip in spherical geometry
    lat0 = np.array([44, 45, 46, 47, 48])
    lon0 = np.array([60, 61, 62, 63])
    epoch0 = 700
    f = lambda x: np.sin(x * np.pi/180.) #value ~sin(lon)
    df = lambda x: np.cos(x * np.pi/180.) * np.pi/180. #derivative ~cos(lon) in units of deg^-1
    v0 = [f(lon0) for _ in range(lat0.shape[0])]
    interp = ki.RegularGridGeospatialInterpolator(value=v0, lat=lat0, lon=lon0, epoch=epoch0)
    dv = interp(lat=[44.5, 46.0], lon=[60.5, 62.0], dlon=1)
    assert dv[0] == pytest.approx(df(60.5), rel=1E-2)
    assert dv[1] == pytest.approx(df(62.0), rel=1E-2)
    dv = interp(lat=[44.5, 46.0], lon=[60.5, 62.0], dlat=1)
    assert dv[0] == pytest.approx(0., abs=1E-6)
    assert dv[1] == pytest.approx(0., abs=1E-6)


def test_irreg_interp_scalar():
    """ Check that we can use the irregular grid interpolator with scalar input """
    v0 = 12.2
    interp = ki.IrregularGridGeospatialInterpolator(value=v0)

    assert interp.coordinates == {}

    # dimensionless interpolation
    v = interp()
    assert v == v0

    # scalar coordinate
    v = interp(lon=1.1)
    assert v == v0

    # multiple scalar coordinates
    v = interp(lon=1.1, epoch=77)
    assert v == v0

    # single coordinate array
    lat = [1, 2]
    v = interp(lat=lat)
    assert v.shape == (len(lat),)
    assert np.all(v == v0)

    # multiple coordinate arrays
    lat = [1, 2]
    lon = [10, 11]
    v = interp(lat=lat, lon=lon)
    assert v.shape == (len(lat),)
    assert np.all(v == v0)

    # if arrays have different sizes, an assertion error is thrown
    with pytest.raises(AssertionError):
        lat = [1, 2]
        lon = [10, 11, 12]
        interp(lat=lat, lon=lon)

    # grid 
    lat = 1
    epoch = [10, 11]
    v = interp(lat=lat, epoch=epoch, grid=True)
    assert v.shape == (1, 2)
    assert np.all(v == v0)

    # another grid 
    depth = [100, 200, 300]    
    epoch = [10, 11]
    v = interp(depth=depth, epoch=epoch, grid=True)
    assert v.shape == (2, 3)
    assert np.all(v == v0)

    # XY interpolation scalar coordinates
    v = interp(x=0, y=0)
    assert v == v0

    # XY interpolation arrays
    v = interp(x=[0, 100], y=[0, -100])
    assert v.shape == (2,)
    assert np.all(v == v0)

    # XY interpolation grid
    v = interp(x=[0, 100], y=[0, -100, 200], epoch=333, grid=True)
    assert v.shape == (2, 3, 1)
    assert np.all(v == v0)


def test_irreg_interp_single_array():
    """ Check that we can use the irregular grid interpolator with single array input """
    v0 = [0, 10, 20]
    lat0 = [0, 1, 2]
    interp = ki.IrregularGridGeospatialInterpolator(value=v0, lat=lat0)

    assert interp.method == "cubic"
    assert np.all(interp.coordinates["lat"] == lat0)
    assert np.all(interp.dims == ["lat"])

    # dimensionless interpolation is not allowed
    with pytest.raises(AssertionError):
        interp()

    # scalar coordinate
    v = interp(lat=1)
    assert v == v0[1]

    # method was downgraded from 'cubic' to 'linear'
    assert interp.method == "linear"

    # multiple scalar coordinates
    v = interp(lat=1, epoch=77)
    assert v == v0[1]

    # linear interpolation is ok
    assert interp.method == "linear"

    # array coordinate
    lat = [1, 2]
    v = interp(lat=lat)
    assert v.shape == (2,)
    assert np.all(v == [10, 20])

    # multiple array coordinates
    lat = [1, 2, 0]
    epoch = [700, 800, 900]
    v = interp(lat=lat, epoch=epoch)
    assert v.shape == (3,)
    assert np.all(v == [10, 20, 0])

    # grid
    lat = [1, 2]
    depth = [22, 42, 62]
    v = interp(lat=lat, depth=depth, grid=True)
    assert v.shape == (2, 3)
    assert np.all(v == [[10, 10, 10], [20, 20, 20]])

    # another grid
    lat = 1
    depth = [22, 42, 62]
    v = interp(lat=lat, depth=depth, grid=True)
    assert v.shape == (1,3)
    assert np.all(v == [[10, 10, 10]])

    # XY interpolation scalar coordinates
    v = interp(x=0, y=0)
    assert v == v0[0]

    # XY interpolation arrays
    one_deg_m = 111195 #approximate distance corresponding to 1 degree change in latitude 
    v = interp(x=[0, 100], y=[0, one_deg_m])
    assert v.shape == (2,)
    assert np.all(np.isclose(v, v0[:2], rtol=1E-3))

    # XY interpolation grid
    v = interp(x=[0, 100], y=[0, one_deg_m, 2*one_deg_m], epoch=333, grid=True)
    assert v.shape == (2, 3, 1)
    assert np.all(np.isclose(v[0, :, 0], v0, rtol=1E-3))


def test_irreg_interp_multiple_arrays():
    """ Check that we can use the irregular grid interpolator with multiple array input """
    lat0 = [0, 1, 2, 0, 1, 2, 0, 1, 2]
    depth0 = [0, 0, 0, 100, 100, 100, 200, 200, 200]
    v0 = [0,  1,  2, 10, 11, 12, 18, 19, 20]
    interp = ki.IrregularGridGeospatialInterpolator(value=v0, lat=lat0, depth=depth0)

    # default method is cubic
    assert interp.method == "cubic"

    assert np.all(interp.coordinates["lat"] == lat0)
    assert np.all(interp.coordinates["depth"] == depth0)
    assert np.all(interp.dims == ["lat", "depth"])

    # dimensionless interpolation is not allowed
    with pytest.raises(AssertionError):
        interp()

    # scalar coordinate
    v = interp(lat=1, depth=100)
    assert isinstance(v, float)
    assert v == pytest.approx(v0[4], rel=1E-6)

    # array coordinates
    v = interp(lat=[1, 2], depth=[100, 200])
    assert v.shape == (2,)
    assert v[0] == pytest.approx(v0[4], rel=1E-6)
    assert v[1] == pytest.approx(v0[8], rel=1E-6)

    # grid
    v = interp(lat=[1, 2], depth=[100, 200, 50], epoch=44, grid=True)
    assert v.shape == (2, 1, 3)
    assert v[0, 0, 0] == pytest.approx(v0[4], rel=1E-6)
    assert v[0, 0, 1] == pytest.approx(v0[7], rel=1E-6)
    assert v[1, 0, 0] == pytest.approx(v0[5], rel=1E-6)
    assert v[1, 0, 1] == pytest.approx(v0[8], rel=1E-6)

    # XY interpolation scalar coordinates
    v = interp(x=0, y=0, depth=0)
    assert v == v0[0]

    # XY interpolation arrays
    one_deg_m = 111195 #approximate distance corresponding to 1 degree change in latitude 
    v = interp(x=[0, 100], y=[0, one_deg_m], depth=[0, 100])
    assert v.shape == (2,)
    assert np.all(np.isclose(v, [v0[0], v0[4]], rtol=1E-3))

    # XY interpolation grid
    v = interp(x=[0, 100], y=[0, one_deg_m, 2*one_deg_m], depth=100, grid=True)
    assert v.shape == (2, 3, 1)
    assert np.all(np.isclose(v[0, :, 0], v0[3:6], rtol=1E-3))


def test_reg_interp_scalar():
    """ Check that we can use the regular grid interpolator with scalar input """
    v0 = 12.2
    interp = ki.RegularGridGeospatialInterpolator(value=v0)

    assert interp.coordinates == {}

    # dimensionless interpolation
    v = interp()
    assert v == v0

    # scalar coordinate
    v = interp(lon=1.1)
    assert v == v0

    # multiple scalar coordinates
    v = interp(lon=1.1, epoch=77)
    assert v == v0

    # single coordinate array
    lat = [1, 2]
    v = interp(lat=lat)
    assert v.shape == (len(lat),)
    assert np.all(v == v0)

    # multiple coordinate arrays
    lat = [1, 2]
    lon = [10, 11]
    v = interp(lat=lat, lon=lon)
    assert v.shape == (len(lat),)
    assert np.all(v == v0)

    # if arrays have different sizes, an assertion error is thrown
    with pytest.raises(AssertionError):
        lat = [1, 2]
        lon = [10, 11, 12]
        interp(lat=lat, lon=lon)

    # grid 
    lat = 1
    epoch = [10, 11]
    v = interp(lat=lat, epoch=epoch, grid=True)
    assert v.shape == (1, 2)
    assert np.all(v == v0)

    # another grid 
    depth = [100, 200, 300]    
    epoch = [10, 11]
    v = interp(depth=depth, epoch=epoch, grid=True)
    assert v.shape == (2, 3)
    assert np.all(v == v0)

    # XY interpolation scalar coordinates
    v = interp(x=0, y=0)
    assert v == v0

    # XY interpolation arrays
    one_deg_m = 111195 #approximate distance corresponding to 1 degree change in latitude 
    v = interp(x=[0, 100], y=[0, one_deg_m])
    assert v.shape == (2,)
    assert np.all(v == v0)

    # XY interpolation grid
    v = interp(x=[0, 100], y=[0, one_deg_m, 2*one_deg_m], epoch=333, grid=True)
    assert v.shape == (2, 3, 1)
    assert np.all(v == v0)


def test_reg_interp_single_array():
    """ Check that we can use the regular grid interpolator with single array input """
    v0 = [0, 10, 20]
    lat0 = [0, 1, 2]
    interp = ki.RegularGridGeospatialInterpolator(value=v0, lat=lat0)

    # default is linear
    assert interp.method == "linear"

    assert np.all(interp.coordinates["lat"] == lat0)
    assert np.all(interp.dims == ["lat"])

    # dimensionless interpolation is not allowed
    with pytest.raises(AssertionError):
        interp()

    # scalar coordinate
    v = interp(lat=1)
    assert v == v0[1]

    # multiple scalar coordinates
    v = interp(lat=1, epoch=77)
    assert v == v0[1]

    # array coordinate
    lat = [1, 2]
    v = interp(lat=lat)
    assert v.shape == (2,)
    assert np.all(v == [10, 20])

    # multiple array coordinates
    lat = [1, 2, 0]
    epoch = [700, 800, 900]
    v = interp(lat=lat, epoch=epoch)
    assert v.shape == (3,)
    assert np.all(v == [10, 20, 0])

    # grid
    lat = [1, 2]
    depth = [22, 42, 62]
    v = interp(lat=lat, depth=depth, grid=True)
    assert v.shape == (2, 3)
    assert np.all(v == [[10, 10, 10], [20, 20, 20]])

    # another grid
    lat = 1
    depth = [22, 42, 62]
    v = interp(lat=lat, depth=depth, grid=True)
    assert v.shape == (1,3)
    assert np.all(v == [[10, 10, 10]])

    # XY interpolation scalar coordinates
    v = interp(x=0, y=0)
    assert v == v0[0]

    # XY interpolation arrays
    one_deg_m = 111195 #approximate distance corresponding to 1 degree change in latitude 
    v = interp(x=[0, 100], y=[0, one_deg_m])
    assert v.shape == (2,)
    assert np.all(np.isclose(v, v0[:2], rtol=1E-3))

    # XY interpolation grid
    v = interp(x=[0, 100], y=[0, one_deg_m, 2*one_deg_m], epoch=333, grid=True)
    assert v.shape == (2, 3, 1)
    assert np.all(np.isclose(v[0, :, 0], v0, rtol=1E-3))


def test_reg_interp_multiple_arrays():
    """ Check that we can use the regular grid interpolator with multiple array input """
    lat0 = [0, 1, 2]
    depth0 = [0, 100, 200]
    v0 = [[ 0,  1,  2],
          [10, 11, 12],
          [18, 19, 20]]
    interp = ki.RegularGridGeospatialInterpolator(value=v0, lat=lat0, depth=depth0)

    # default method is linear
    assert interp.method == "linear"

    assert np.all(interp.coordinates["lat"] == lat0)
    assert np.all(interp.coordinates["depth"] == depth0)
    assert np.all(interp.dims == ["lat", "depth"])

    # dimensionless interpolation is not allowed
    with pytest.raises(AssertionError):
        interp()

    # scalar coordinate
    v = interp(lat=1, depth=100)
    assert isinstance(v, float)
    assert v == pytest.approx(v0[1][1], rel=1E-6)

    # array coordinates
    v = interp(lat=[1, 2], depth=[100, 200])
    assert v.shape == (2,)
    assert v[0] == pytest.approx(v0[1][1], rel=1E-6)
    assert v[1] == pytest.approx(v0[2][2], rel=1E-6)

    # grid
    v = interp(lat=[1, 2], depth=[100, 200, 50], epoch=44, grid=True)
    assert v.shape == (2, 1, 3)
    assert v[0, 0, 0] == pytest.approx(v0[1][1], rel=1E-6)
    assert v[0, 0, 1] == pytest.approx(v0[1][2], rel=1E-6)
    assert v[1, 0, 0] == pytest.approx(v0[2][1], rel=1E-6)
    assert v[1, 0, 1] == pytest.approx(v0[2][2], rel=1E-6)

    # XY interpolation scalar coordinates
    v = interp(x=0, y=0, depth=0)
    assert v == v0[0][0]

    # XY interpolation arrays
    one_deg_m = 111195 #approximate distance corresponding to 1 degree change in latitude 
    v = interp(x=[0, 100], y=[0, one_deg_m], depth=[0, 100])
    assert v.shape == (2,)
    assert np.all(np.isclose(v, [v0[0][0], v0[1][1]], rtol=1E-3))

    # XY interpolation grid
    v = interp(x=[0, 100], y=[0, one_deg_m, 2*one_deg_m], depth=100, grid=True)
    assert v.shape == (2, 3, 1)
    assert np.all(np.isclose(v[0, :, 0], np.array(v0)[:,1], rtol=1E-3))


def test_reg_interp_lat_lon_arrays():
    """ Check that we can use the regular grid interpolator on lat-lon input """
    lat0 = np.array([44, 45, 46, 47, 48])
    lon0 = np.array([60, 61, 62, 63])
    epoch0 = 700
    v0 = np.random.rand(len(lat0), len(lon0))

    interp = ki.RegularGridGeospatialInterpolator(value=v0, lat=lat0, lon=lon0, epoch=epoch0)

    assert isinstance(interp._interp, ki._RectSphereBivariateSpline)
    assert interp.method == "spline"

    assert np.all(interp.coordinates["lat"] == lat0)
    assert np.all(interp.coordinates["lon"] == lon0)
    assert np.all(interp.dims == ["lat","lon"])

    # multiple array coordinates
    lat = [44, 45, 46]
    lon = [60, 61, 60]
    v = interp(lat=lat, lon=lon)
    assert v.shape == (3,)
    assert v[0] == pytest.approx(v0[0,0], rel=1E-6)
    assert v[1] == pytest.approx(v0[1,1], rel=1E-6)
    assert v[2] == pytest.approx(v0[2,0], rel=1E-6)

    # grid
    lat = [44, 45, 46]
    lon = [60, 61]
    v = interp(lat=lat, lon=lon, grid=True)
    assert v.shape == (3, 2)
    assert np.all(np.isclose(v, v0[:3, :2], rtol=1E-6))
    v = interp(lat=lat, lon=lon, depth=[321, 1321], grid=True)
    assert v.shape == (3, 2, 2)
    assert np.all(np.isclose(v[:, :, 0], v0[:3, :2], rtol=1E-6))

    # XY interpolation scalar coordinates
    v = interp(x=0, y=0, depth=0, origin=(45, 61))
    assert v == pytest.approx(v0[1][1], rel=1E-6)

    # XY interpolation arrays
    one_deg_m = 111195 #approximate distance corresponding to 1 degree change in latitude 
    v = interp(x=[0, 0], y=[0, one_deg_m], depth=[0, 100], origin=(46, 61))
    assert v.shape == (2,)
    assert np.all(np.isclose(v, [v0[2][1], v0[3][1]], rtol=1E-3))

    # XY interpolation grid
    v = interp(x=[0, 100], y=[0, one_deg_m, 2*one_deg_m], depth=100, origin=(46, 61), grid=True)
    assert v.shape == (2, 3, 1)
    assert np.all(np.isclose(v[0, :, 0], np.array(v0)[2:,1], rtol=1E-3))


def test_create_regular_grid():
    """ Check that the _create_regular_grid helper function returns grids of expected shape and range
    """
    # coordinates provide 50% coverage of rectangle with corners (0,10), (2,10), (2,210), (0,210)
    lat = [1, 2, 1, 0, 1]
    depth = [10, 110, 210, 110, 110]

    # regular grid with specified shape
    g, c = ki._create_regular_grid(
        lat=lat, depth=depth, grid_shape={"lat":11, "depth":11}, return_coverage=True
    )
    assert np.all(g["lat"] == np.linspace(0, 2, 11))
    assert np.all(g["depth"] == np.linspace(10, 210, 11))
    assert c == 0.5

    # regular grid with specified bin size
    g, c = ki._create_regular_grid(
        lat=lat, depth=depth, bin_size={"lat":0.5, "depth":40}, return_coverage=True
    )
    assert np.all(g["lat"] == np.linspace(0, 2, 5))
    assert np.all(g["depth"] == np.linspace(10, 210, 6))
    assert c == 0.5

    # regular grid with bins automatically inferred
    g, c = ki._create_regular_grid(
        lat=lat, depth=depth, return_coverage=True
    )
    assert np.all(g["lat"] == np.linspace(0, 2, 3))
    assert np.all(g["depth"] == np.linspace(10, 210, 3))
    assert c == 0.5


def test_get_interpolator():
    """ Check that the get_interpolator convenience method returns interpolators as it should"""
    # data on a regular grid
    lat0 = np.array([44, 45, 46, 47, 48])
    lon0 = np.array([60, 61, 62, 63])
    epoch0 = 700
    v0 = np.random.rand(len(lat0), len(lon0))
    interp = ki.get_interpolator(value=v0, lat=lat0, lon=lon0, epoch=epoch0)
    assert isinstance(interp, ki.RegularGridGeospatialInterpolator)
    assert interp.method == "spline"
    assert np.all(interp.coordinates["lat"] == lat0)
    assert np.all(interp.coordinates["lon"] == lon0)
    assert np.all(interp.dims == ["lat","lon"])

    # data on a irregular grid
    lat0 = [0, 1, 2, 0, 1, 2, 0, 1, 2]
    depth0 = [0, 0, 0, 100, 100, 100, 200, 200, 200]
    v0 = [0,  1,  2, 10, 11, 12, 18, 19, 20]
    interp = ki.get_interpolator(value=v0, lat=lat0, depth=depth0)
    assert isinstance(interp, ki.RegularGridGeospatialInterpolator)
    assert interp.method == "linear" #default
    assert np.all(interp.coordinates["lat"] == np.unique(lat0))
    assert np.all(interp.coordinates["depth"] == np.unique(depth0))
    assert np.all(interp.dims == ["lat", "depth"])


"""
    =====================================================
    |                                                   |
    |   Legacy (kadlu < v2.4.0) tests below this point  |
    |                                                   | 
    =====================================================
"""

def test_irregular_interp_2d():
    """Check that IrregularGridGeospatialInterpolator behaves as expected for (very simple) 2D data"""
    # check that we can interpolate a single data point
    lat = [1.]
    lon = [2.]
    v = [3.]
    itp = ki.IrregularGridGeospatialInterpolator(value=v,lat=lat,lon=lon,method="linear")
    assert itp.method == None
    # specifying x,y as lists, we get an array back
    w = itp(lat=lat, lon=lon)
    assert w.shape == (1,)
    assert w == v
    # returned array has expected shape when grid = True
    w = itp(lat=[1,3], lon=[2,4,5], grid=True)
    assert w.shape == (2,3)
    assert np.all(w == 3.)
    # check that we can interpolate two data points
    lat = [1., 2.]
    lon = [2., 4.]
    v = [3., 6.]
    itp = ki.IrregularGridGeospatialInterpolator(value=v,lat=lat,lon=lon,method="linear")
    assert itp.method == "linear"
    w = itp(lat=lat[0], lon=lon[0])
    # check that method was downgraded to 'nearest' after first call
    assert itp.method == "nearest"
    assert w == v[0]
    w = itp(lat=[1,3], lon=[2,3,5], grid=True)
    assert w.shape == (2,3)
    assert w[0,0] == 3. #1,2 -> 1,2
    assert w[0,1] == 3. #1,3 -> 1,2
    assert w[0,2] == 6. #1,5 -> 2,4 
    assert w[1,0] == 3. or w[1,0] == 6. #3,2 -> 1,2 or 2,4
    assert w[1,1] == 6. #3,3 -> 2,4
    assert w[1,2] == 6. #3,5 -> 2,4
    # check that NaN values get replaced with nearest valid value
    lat = [1., 2., 0., -3.]
    lon = [2., 4., -2., 2.]
    v = [3., 6., 12., 9.]
    itp = ki.IrregularGridGeospatialInterpolator(value=v,lat=lat,lon=lon,method="linear")
    assert itp.method == "linear"
    # interpolate outside the convex hull
    w = itp(lat=2, lon=4.1, replace_nan=False)
    # check that method is unchanged by call
    assert itp.method == "linear"
    assert np.isnan(w)
    w = itp(lat=2, lon=4.1)
    assert w == 6.


def test_irregular_interp_3d():
    """Check that IrregularGridGeospatialInterpolator behaves as expected for 3D data"""
    # create fake data
    lat = np.linspace(0., 3., 4) #[0,1,2,3]
    lon = np.linspace(0., 4., 5) #[0,1,2,3,4]
    depth = np.linspace(0., 100., 3) #[0,50,100]
    lat, lon, depth = np.meshgrid(lat, lon, depth, indexing="ij")
    lat = lat.flatten()
    lon = lon.flatten()
    depth = depth.flatten()
    temp = np.ones((4, 5, 3))
    for i in range(3):
        temp[:, :, i] *= i  # temperature increases with depth: 0,1,2
    for i in range(5):
        temp[:, i, :] *= i  # temperature increases with longitude 0,0,0,0,0; 0,1,2,3,4; 0,2,4,6,8
    
    temp = temp.flatten()
    
    # initialize interpolator
    ip = ki.IrregularGridGeospatialInterpolator(value=temp, lat=lat, lon=lon, depth=depth, method="linear")

    # --- 4 latitudes ---
    lats = [1.2, 2.2, 0.2, 1.55]
    # --- 4 longitudes ---
    lons = [0.0, 1.0, 2.0, 4.0]
    # --- 4 depths ---
    depths = [0, 25, 50, 150]

    # interpolate all at once
    temps = ip(lat=lats, lon=lons, depth=depths)
    assert temps[0] == pytest.approx(0.0, abs=0.01)
    assert temps[1] == pytest.approx(0.5, abs=0.01)
    assert temps[2] == pytest.approx(2.0, abs=0.01)
    assert np.all(np.logical_and(temps >= -5, temps <= 105))

    # interpolate one at a time
    ti = list()
    for lat, lon, depth in zip(lats, lons, depths):
        ti.append(ip(lat=lat, lon=lon, depth=depth))

    # check that the all-at-once and one-at-a-time
    # approaches give the same result
    for tii, t in zip(ti, temps):
        assert tii == pytest.approx(t, rel=1e-3)


def test_interp_2x2_grid():
    values = np.array([[0, 2], [0, 2]])
    lats = np.array([0, 1])
    lons = np.array([0, 1])
    ip = ki.get_interpolator(value=values, lat=lats, lon=lons, method="linear")
    assert ip(lat=0, lon=0) == 0
    assert ip(lat=1, lon=1) == 2
    res = ip(lat=0, lon=0.5)
    assert np.abs(res - 1.) < 1e-6
    res = ip(lat=[0.5, 0.5], lon=[0.2, 0.3])
    assert np.all(np.abs(res - [0.4, 0.6]) < 1e-6)
    res = ip(lat=[0.5, 0.6, 0.7], lon=[0.2, 0.3], grid=True)
    assert res.shape == (3, 2)
    assert np.all(np.abs(res - [0.4, 0.6]) < 1e-6)


def test_interp_1x1_grid():
    values = np.array([[17]])
    lats = np.array([0])
    lons = np.array([0])
    ip = ki.get_interpolator(value=values, lat=lats, lon=lons)
    assert ip(lat=0, lon=0) == 17
    assert ip(lat=1, lon=1) == 17
    v = ip(lat=[1,15,-1], lon=[1,200], grid=True)
    assert v.shape == (3, 2)
    assert np.all(v == 17)


def test_interp_1x2_grid():
    values = np.array([[0, 2]])
    lats = np.array([0])
    lons = np.array([0, 1])
    ip = ki.get_interpolator(value=values, lat=lats, lon=lons)
    assert ip(lat=0, lon=0) == 0
    assert ip(lat=1, lon=1) == 2
    res = ip(lat=0, lon=0.5)
    assert np.abs(res - 1.) < 1e-6


def test_interpolate_bathymetry_using_latlon_coordinates():

    # load bathy data
    path = path_to_assets + '/bornholm.mat'
    bathy, lat, lon = load_data_from_file(path)

    # initialize interpolator
    ip = ki.get_interpolator(value=bathy, lat=lat, lon=lon)

    # interpolate at a single point on the grid
    z = ip(lat=lat[0], lon=lon[0])
    z = int(z)
    assert z == bathy[0, 0]

    # interpolate at a single point between two grid points
    x = (lat[1] + lat[2]) / 2
    z = ip(lat=x, lon=lon[0])
    z = float(z)
    zmin = min(bathy[1, 0], bathy[2, 0])
    zmax = max(bathy[1, 0], bathy[2, 0])
    assert z >= zmin
    assert z <= zmax

    # interpolate at two points
    x1 = (lat[1] + lat[2]) / 2
    x2 = (lat[2] + lat[3]) / 2
    z = ip(lat=[x1, x2], lon=lon[0], grid=True)
    zmin = min(bathy[1, 0], bathy[2, 0])
    zmax = max(bathy[1, 0], bathy[2, 0])
    assert z[0] >= zmin
    assert z[0] <= zmax
    zmin = min(bathy[2, 0], bathy[3, 0])
    zmax = max(bathy[2, 0], bathy[3, 0])
    assert z[1] >= zmin
    assert z[1] <= zmax

    # interpolate with grid = True/False
    x1 = (lat[1] + lat[2]) / 2
    x2 = (lat[2] + lat[3]) / 2
    y1 = (lon[1] + lon[2]) / 2
    y2 = (lon[2] + lon[3]) / 2
    z = ip(lat=[x1, x2], lon=[y1, y2], grid=False)
    assert np.ndim(z) == 1
    assert z.shape[0] == 2
    z = ip(lat=[x1, x2], lon=[y1, y2], grid=True)
    assert np.ndim(z) == 2
    assert z.shape[0] == 2
    assert z.shape[1] == 2


def test_interpolation_grids_are_what_they_should_be():
    # load data and initialize interpolator
    path = path_to_assets + '/bornholm.mat'
    _, lat, lon = load_data_from_file(path)
    lat_c = 0.5 * (lat[0] + lat[-1])
    lon_c = 0.5 * (lon[0] + lon[-1])
    assert lat_c, lon_c == center_point(lat, lon)


def test_interpolation_tables_agree_on_latlon_grid():
    # load data and initialize interpolator
    path = path_to_assets + '/bornholm.mat'
    bathy, lat, lon = load_data_from_file(path)
    ip = ki.get_interpolator(bathy, lat=lat, lon=lon)

    # lat fixed
    ilat = int(len(ip.coordinates["lat"]) / 2)
    lat = ip.coordinates["lat"][ilat]
    for lon in ip.coordinates["lon"]:
        bll = ip(lat=lat, lon=lon)
        x, y = LLtoXY(
            lat=lat,
            lon=lon,
            lat_ref=ip.origin[0],
            lon_ref=ip.origin[1]
        )
        bxy = ip(x=x, y=y)
        assert bxy == pytest.approx(bll, rel=1e-3) or bxy == pytest.approx(bll, abs=0.1)

    # lon fixed
    ilon = int(len(ip.coordinates["lon"]) / 2)
    lon = ip.coordinates["lon"][ilon]
    for lat in ip.coordinates["lat"]:
        bll = ip(lat=lat, lon=lon)
        x, y = LLtoXY(
            lat=lat,
            lon=lon,
            lat_ref=ip.origin[0],
            lon_ref=ip.origin[1]
        )
        bxy = ip(x=x, y=y)
        assert bxy == pytest.approx(bll, rel=1e-3) or bxy == pytest.approx(bll, abs=0.1)


def test_interpolation_tables_agree_anywhere():
    # load data and initialize interpolator
    path = path_to_assets + '/bornholm.mat'
    bathy, lat, lon = load_data_from_file(path)
    ip = ki.get_interpolator(bathy, lat=lat, lon=lon)

    # --- at origo ---
    lat_c, lon_c = center_point(lat, lon)
    #lat_c = ip.origin.latitude
    #lon_c = ip.origin.longitude
    z_ll = ip(lat=lat_c, lon=lon_c)  # interpolate using lat-lon
    z_ll = float(z_ll)
    z_xy = ip(x=0, y=0)  # interpolate using x-y
    z_xy = float(z_xy)
    assert z_ll == pytest.approx(z_xy, rel=1e-3) or z_xy == pytest.approx(z_ll, abs=0.1)

    # --- 0.1 degrees north of origo ---
    lat = lat_c + 0.1
    lon = lon_c
    x, y = LLtoXY(lat=lat, lon=lon, lat_ref=lat_c, lon_ref=lon_c)
    z_ll = ip(lat=lat, lon=lon)
    z_ll = float(z_ll)
    z_xy = ip(x=x, y=y)
    z_xy = float(z_xy)
    assert z_ll == pytest.approx(z_xy, rel=1e-3) or z_xy == pytest.approx(z_ll, abs=0.1)

    # --- 0.08 degrees south of origo ---
    lat = lat_c - 0.08
    lon = lon_c
    x, y = LLtoXY(lat=lat, lon=lon, lat_ref=lat_c, lon_ref=lon_c)
    z_ll = ip(lat=lat, lon=lon)
    z_ll = float(z_ll)
    z_xy = ip(x=x, y=y)
    z_xy = float(z_xy)
    assert z_ll == pytest.approx(z_xy, rel=1e-3) or z_xy == pytest.approx(z_ll, abs=0.1)

    # --- at shifted origo ---
    bathy, lat, lon = load_data_from_file(path)
    ip = ki.get_interpolator(bathy, lat=lat, lon=lon, origin=(55.30, 15.10))
    lat_c = ip.origin[0]
    lon_c = ip.origin[1]
    z_ll = ip(lat=lat_c, lon=lon_c)  # interpolate using lat-lon
    z_ll = float(z_ll)
    z_xy = ip(x=0, y=0)  # interpolate using x-y
    z_xy = float(z_xy)
    assert z_ll == pytest.approx(z_xy, rel=1e-3) or z_xy == pytest.approx(z_ll, abs=0.1)


def test_interpolation_tables_agree_on_ll_grid_for_dbarclays_data():
    # load data and initialize interpolator
    path = path_to_assets + '/BathyData_Mariana_500kmx500km.mat'
    bathy, lat, lon = load_data_from_file(
        path,
        lat_name='latgrat',
        lon_name='longrat',
        val_name='mat',
        lon_axis=0
    )
    ip = ki.get_interpolator(bathy, lat=lat, lon=lon)

    # lat fixed
    ilat = int(len(ip.coordinates["lat"]) / 2)
    lat = ip.coordinates["lat"][ilat]
    for lon in ip.coordinates["lon"]:
        bll = ip(lat=lat, lon=lon)
        x, y = LLtoXY(
            lat=lat,
            lon=lon,
            lat_ref=ip.origin[0],
            lon_ref=ip.origin[1]
        )
        bxy = ip(x=x, y=y)
        assert bxy == pytest.approx(bll, rel=1e-3) or bxy == pytest.approx(bll, abs=0.1)

    # lon fixed
    ilon = int(len(ip.coordinates["lon"]) / 2)
    lon = ip.coordinates["lon"][ilon]
    for lat in ip.coordinates["lat"]:
        bll = ip(lat=lat, lon=lon)
        x, y = LLtoXY(
            lat=lat,
            lon=lon,
            lat_ref=ip.origin[0],
            lon_ref=ip.origin[1]
        )
        bxy = ip(x=x, y=y)
        assert bxy == pytest.approx(bll, rel=1e-3) or bxy == pytest.approx(bll, abs=0.1)


def test_interpolation_tables_agree_anywhere_for_dbarclays_data():
    # load data and initialize interpolator
    path = path_to_assets + '/BathyData_Mariana_500kmx500km.mat'
    bathy, lat, lon = load_data_from_file(
        path,
        lat_name='latgrat',
        lon_name='longrat',
        val_name='mat',
        lon_axis=0
    )
    ip = ki.get_interpolator(bathy, lat=lat, lon=lon)

    # --- at origo ---
    lat_c = ip.origin[0]
    lon_c = ip.origin[1]
    z_ll = ip(lat=lat_c, lon=lon_c)  # interpolate using lat-lon
    z_ll = float(z_ll)
    z_xy = ip(x=0, y=0)  # interpolate using x-y
    z_xy = float(z_xy)
    assert z_ll == pytest.approx(z_xy, rel=1E-3) or z_ll == pytest.approx(z_xy, abs=0.1)

    # --- at shifted origo ---
    bathy, lat, lon = load_data_from_file(
        path,
        lat_name='latgrat',
        lon_name='longrat',
        val_name='mat',
        lon_axis=0
    )
    ip = ki.get_interpolator(bathy, lat=lat, lon=lon, origin=(9., 140.))
    lat_c = ip.origin[0]
    lon_c = ip.origin[1]
    z_ll = ip(lat=lat_c, lon=lon_c)  # interpolate using lat-lon
    z_ll = float(z_ll)
    z_xy = ip(x=0, y=0)  # interpolate using x-y
    z_xy = float(z_xy)
    assert z_ll == pytest.approx(z_xy, rel=1E-3) or z_ll == pytest.approx(z_xy, abs=0.1)


def test_mariana_trench_is_in_correct_location():
    # load data and initialize interpolator
    path = path_to_assets + '/BathyData_Mariana_500kmx500km.mat'
    bathy, lat, lon = load_data_from_file(
        path,
        lat_name='latgrat',
        lon_name='longrat',
        val_name='mat',
        lon_axis=0
    )
    ip = ki.get_interpolator(bathy, lat=lat, lon=lon)
    d = ip(lat=11.3733, lon=142.5917)
    assert d < -10770
    d = ip(lat=12.0, lon=142.4)
    assert d > -3000
    d = ip(lat=11.4, lon=143.1)
    assert d < -9000


def test_can_interpolate_multiple_points_in_ll():
    # load data and initialize interpolator
    path = path_to_assets + '/bornholm.mat'
    bathy, lat, lon = load_data_from_file(path)
    ip = ki.get_interpolator(bathy, lat=lat, lon=lon)
    # coordinates of origin
    lat_c = ip.origin[0]
    lon_c = ip.origin[1]
    # --- 4 latitudes ---
    lats = [lat_c, lat_c + 0.1, lat_c - 0.2, lat_c + 0.03]
    # --- 4 longitudes ---
    lons = [lon_c, lon_c + 0.15, lon_c - 0.08, lon_c - 0.12]
    # interpolate all at once
    values_all = ip(lat=lats, lon=lons)
    # interpolate one at a time
    values_one = list()
    for lat, lon in zip(lats, lons):
        values_one.append(ip(lat=lat, lon=lon))
    # check that the values agree
    for a,b in zip(values_one, values_all):
        assert a == pytest.approx(b, rel=1e-3)
    # interpolate on a grid
    values_grid = ip(lat=lats, lon=lons, grid=True)
    # check that diagonal values agree
    for i,v in enumerate(values_all):
        assert values_grid[i,i] == pytest.approx(v, rel=1e-3)


def test_can_interpolate_multiple_points_in_xy():
    # load data and initialize interpolator
    path = path_to_assets + '/bornholm.mat'
    bathy, lat, lon = load_data_from_file(path)
    ip = ki.get_interpolator(bathy, lat=lat, lon=lon)
    # --- 4 x coordinates ---
    xs = [0, 1000, -2000, 300]
    # --- 4 y coordinates ---
    ys = [0, 1500, 800, -120]
    # interpolate all at once
    values_all = ip(x=xs, y=ys)
    # interpolate one at a time
    values_one = list()
    for x, y in zip(xs, ys):
        values_one.append(ip(x=x, y=y))
    # check that the values agree
    for a,b in zip(values_one, values_all):
        assert a == pytest.approx(b, rel=1e-3)
    # interpolate on a grid
    values_grid = ip(x=xs, y=ys, grid=True)
    # check that diagonal values agree
    for i,v in enumerate(values_all):
        assert values_grid[i,i] == pytest.approx(v, rel=1e-3)


def test_can_interpolate_regular_grid():
    # create fake data
    lat = np.array([44, 45, 46, 47, 48])
    lon = np.array([60, 61, 62, 63])
    bathy = np.random.rand(len(lat), len(lon))
    # initialize interpolator
    ip = ki.get_interpolator(bathy, lat=lat, lon=lon)
    # check value at grid point
    b = ip(lat=45, lon=62)
    assert b == pytest.approx(bathy[1, 2], abs=1E-9)


def test_can_interpolate_irregular_grid():
    # create fake data
    lat = np.array([0.0, 1.0, 1.5, 2.1, 3.0])
    lon = np.array([0.0, 2.0, 0.2, 0.7, 1.2])
    bathy = np.array([-90.0, -200.0, -140.0, -44.0, -301.0])
    # initialize interpolator
    ip = ki.get_interpolator(bathy, lat=lat, lon=lon)
    # --- 4 latitudes ---
    lats = [0.01, 1.0, 0.5, 2.1]
    # --- 4 longitudes ---
    lons = [0.01, 2.0, 1.0, 0.71]
    # interpolate all at once
    depths = ip(lat=lats, lon=lons)
    assert depths[1] == pytest.approx(-200, abs=0.1)
    assert depths[2] < -90 and depths[2] > -200
    # interpolate one at a time
    zi = list()
    for lat, lon in zip(lats, lons):
        zi.append(ip(lat=lat, lon=lon))
    # check that the all-at-once and one-at-a-time
    # approaches give the same result
    for z, d in zip(zi, depths):
        assert z == pytest.approx(d, rel=1e-3)


def test_can_interpolate_irregular_3d_grid():
    # create fake data
    lat = np.linspace(0., 3., 4)
    lon = np.linspace(0., 4., 5)
    depth = np.linspace(0., 100., 3)
    lat, lon, depth = np.meshgrid(lat, lon, depth, indexing="ij")
    lat = lat.flatten()
    lon = lon.flatten()
    depth = depth.flatten()
    temp = np.ones((4, 5, 3))
    for i in range(3):
        temp[:, :, i] *= i  # temperature increases with depth
    for i in range(5):
        temp[:, i, :] *= i  # temperature increases with longitude
    
    temp = temp.flatten()
    
    # initialize interpolator
    ip = ki.get_interpolator(temp, lat=lat, lon=lon, depth=depth)

    # --- 4 latitudes ---
    lats = [1.2, 2.2, 0.2, 1.55]
    # --- 4 longitudes ---
    lons = [0.0, 1.0, 2.0, 4.0]
    # --- 4 depths ---
    depths = [0, 25, 50, 150]
    # interpolate all at once
    temps = ip(lat=lats, lon=lons, depth=depths)
    assert temps[0] == pytest.approx(0.0, abs=0.01)
    assert temps[1] == pytest.approx(0.5, abs=0.01)
    assert temps[2] == pytest.approx(2.0, abs=0.01)
    assert np.all(np.logical_and(temps >= -5, temps <= 105))
    # interpolate one at a time
    ti = list()
    for lat, lon, depth in zip(lats, lons, depths):
        ti.append(ip(lat=lat, lon=lon, depth=depth))
    # check that the all-at-once and one-at-a-time
    # approaches give the same result
    for tii, t in zip(ti, temps):
        assert tii == pytest.approx(t, rel=1e-3)


def test_can_interpolate_irregular_3d_grid_where_one_dimension_only_has_1_point():
    # create fake data
    lat = np.linspace(0., 3., 4)
    lon = np.linspace(0., 4., 5)
    depth = np.linspace(0., 100., 1)
    lat, lon, depth = np.meshgrid(lat, lon, depth, indexing="ij")
    lat = lat.flatten()
    lon = lon.flatten()
    depth = depth.flatten()
    temp = np.ones((4, 5, 1))
    for i in range(5):
        temp[:, i, :] *= i  # temperature increases with longitude
    
    temp = temp.flatten()
    
    # initialize interpolator
    ip = ki.get_interpolator(temp, lat=lat, lon=lon, depth=depth)

    # --- 4 latitudes ---
    lats = [1.2, 2.2, 0.2, 1.55]
    # --- 4 longitudes ---
    lons = [0.0, 1.0, 2.0, 4.0]
    # --- 4 depths ---
    depths = [0, 25, 50, 150]
    # interpolate all at once
    temps = ip(lat=lats, lon=lons, depth=depths)
    assert temps[0] == pytest.approx(0.0, abs=0.01)
    assert temps[1] == pytest.approx(1.0, abs=0.01)
    assert temps[2] == pytest.approx(2.0, abs=0.01)
    assert np.all(np.logical_and(temps >= -5, temps <= 105))
    # interpolate one at a time
    ti = list()
    for lat, lon, depth in zip(lats, lons, depths):
        ti.append(ip(lat=lat, lon=lon, depth=depth))
    # check that the all-at-once and one-at-a-time
    # approaches give the same result
    for tii, t in zip(ti, temps):
        assert tii == pytest.approx(t, rel=1e-3)


def test_can_interpolate_gebco_data():
    # load data and initialize interpolator
    south, west = 43, -60
    north, east = 44, -59
    bathy, lat, lon = kadlu.load(
        var='bathymetry',
        source='gebco',
        south=south,
        north=north,
        west=west,
        east=east
    )

    ip = ki.get_interpolator(bathy, lat=lat, lon=lon, method='nearest')
    # --- 4 latitudes ---
    lats = [43.3, 43.2, 43.7, 43.5]
    # --- 4 longitudes ---
    lons = [-59.6, -59.8, -59.2, -59.3]
    # interpolate
    depths = ip(lat=lats, lon=lons)
    zi = list()
    for lat, lon in zip(lats, lons):
        zi.append(ip(lat=lat, lon=lon))
    for z, d in zip(zi, depths):
        assert z == pytest.approx(d, rel=1e-3)
    # interpolate on grid
    depths_grid = ip(lat=lats, lon=lons, grid=True)
    assert depths_grid.shape[0] == 4
    assert depths_grid.shape[1] == 4
    for i in range(4):
        assert depths_grid[i, i] == depths[i]


def test_interpolate_uniform_3d_data():
    N = 10
    np.random.seed(1)
    # create fake data
    val = np.ones(shape=(N, N, N)) 
    lat = np.arange(N)
    lon = np.arange(N)
    depth = np.arange(N)
    # initialize interpolator
    ip = ki.get_interpolator(val, lat=lat, lon=lon, depth=depth)
    # check interpolation at a few random points
    lats = np.random.rand(3) * (N - 1)
    lons = np.random.rand(3) * (N - 1)
    depths = np.random.rand(3) * (N - 1)
    vi = ip(lat=lats, lon=lons, depth=depths)
    for v in vi:
        assert v == pytest.approx(1, abs=1E-9)
    # check interpolation on a grid
    lats = np.random.rand(3) * (N - 1)
    lons = np.random.rand(4) * (N - 1)
    depths = np.random.rand(5) * (N - 1)
    vi = ip(lat=lats, lon=lons, depth=depths, grid=True)
    assert vi.shape[0] == 3
    assert vi.shape[1] == 4
    assert vi.shape[2] == 5
    assert np.all(np.abs(vi - 1.0) < 1E-9)


def test_interpolate_3d_data_with_constant_slope():
    N = 10
    np.random.seed(1)
    # create fake data
    val = np.ones(shape=(N, N, N))
    for k in range(N):
        val[:, :, k] = k * val[:, :, k]
    lat = np.arange(N)
    lon = np.arange(N)
    depth = np.arange(N)
    # initialize interpolator
    ip = ki.get_interpolator(val, lat=lat, lon=lon, depth=depth)
    # check interpolation
    lats = np.array([4, 4, 4])
    lons = np.array([4, 4, 4])
    depths = np.array([4, 4.5, 5])
    vi = ip(lat=lats, lon=lons, depth=depths)
    assert vi[0] == pytest.approx(4, abs=1E-9)
    assert vi[1] == pytest.approx(4.5, abs=1E-9)
    assert vi[2] == pytest.approx(5, abs=1E-9)
    # check interpolation on grid
    lats = np.array([2, 3, 4])
    lons = np.array([1, 2, 3])
    depths = np.array([4, 4.5, 5])
    vi = ip(lat=lats, lon=lons, depth=depths, grid=True)
    assert np.all(np.abs(vi[:, :, 0] - 4) < 1E-9)
    assert np.all(np.abs(vi[:, :, 1] - 4.5) < 1E-9)
    assert np.all(np.abs(vi[:, :, 2] - 5) < 1E-9)


def test_interpolate_3d_data_using_xy_coordinates():
    N = 11
    np.random.seed(1)
    # create fake data
    val = np.ones(shape=(N, N, N))
    for k in range(N):
        val[:, :, k] = k * val[:, :, k]
    lat = np.arange(N)
    lon = np.arange(N)
    depth = np.arange(N)
    # initialize interpolator
    ip = ki.get_interpolator(val, lat=lat, lon=lon, depth=depth)
    # check interpolation
    x = np.array([0, 100, 200])
    y = np.array([0, 100, 200])
    depths = np.array([4, 4.5, 5])
    vi = ip(x=x, y=y, depth=depths)
    assert vi[0] == pytest.approx(4, abs=1E-9)
    assert vi[1] == pytest.approx(4.5, abs=1E-9)
    assert vi[2] == pytest.approx(5, abs=1E-9)


def test_interpolate_3d_outside_grid():
    N = 10
    np.random.seed(1)
    # create fake data
    val = np.ones(shape=(N, N, N))
    lat = np.arange(N)
    lon = np.arange(N)
    depth = np.arange(N)
    # initialize interpolator
    ip = ki.get_interpolator(val, lat=lat, lon=lon, depth=depth)
    # check interpolation outside grid
    lats = 20
    lons = 5
    depths = 5
    vi = ip(lat=lats, lon=lons, depth=depths)
    assert vi == pytest.approx(1, abs=1E-3)


def test_interpolate_uniform_2d():
    ip = ki.get_interpolator(17)
    v = ip(lat=5, lon=2000)
    assert v == 17
    v = ip(lat=[5, 12, 13], lon=[2000, 0, 1])
    assert np.all(v == 17)
    assert v.shape[0] == 3
    v = ip(lat=[5, 12, 13], lon=[2000, 0], grid=True)
    assert np.all(v == 17)
    assert v.shape[0] == 3
    assert v.shape[1] == 2


def test_interpolate_uniform_3d():
    ip = ki.get_interpolator(17)
    v = ip(lat=5, lon=2000, depth=-10)
    assert v == 17
    v = ip(lat=[5, 12, 13], lon=[2000, 0, 1], depth=[0, 2, -3])
    assert np.all(v == 17)
    assert v.shape[0] == 3
    v = ip(lat=[5, 12, 13], lon=[2000, 0], depth=[0, 2, -3], grid=True)
    assert np.all(v == 17)
    assert v.shape[0] == 3
    assert v.shape[1] == 2
    assert v.shape[2] == 3


def test_interpolate_depth_3d():
    ip = ki.get_interpolator(
        value=[0, 1, 4, 9],
        depth=[0, 1, 2, 3],
        method='cubic'
    )
    # inside range
    v = ip(lat=5, lon=2000, depth=1.5)
    assert v == pytest.approx(1.5 * 1.5, rel=1E-4)
    # outside range
    v = ip(lat=5, lon=2000, depth=3.5)
    assert v == pytest.approx(3.5 * 3.5, rel=1E-4)



import logging
import numpy as np
import traceback
import numpy as np
from tqdm import tqdm
from scipy.spatial import ConvexHull
from scipy.spatial._qhull import QhullError
from scipy.interpolate import griddata, RectSphereBivariateSpline, RegularGridInterpolator
from kadlu.utils import as_array, center_point, reverse_index_map, torad, deg2rad, XYtoLL, DLDL_over_DXDY


'''
    TODO: 
     [ ] complete docstrings
     [ ] rescale depth values relative to lat/lon so that "nearest" point is nearest in the sense of euclidian distance ?
'''


""" 
    Names and ordering of dimensions used by all interpolation classes in Kadlu
"""
GEOSPATIAL_DIMS = ["lat", "lon", "epoch", "depth"]


def get_interpolator(value, **kwargs):
    """ Convenience function for initialising a interpolator for a given set of data.

        Args:
            value: array-like
                Data values

        Keyword args:                
            lat: array-like
                Latitudes in degrees. 
            lon: array-like
                Longitudes in degrees. 
            depth: array-like
                Depths in meters below the sea surface. 
            epoch: array-like
                Time in hours since 2000-01-01 00:00:00. 
            origin: tuple(float,float)
                Reference location used as the origo of XY coordinate system. If not specified, 
                the center point of the lat,lon coordinates will be used.
            method: str
                Preferred interpolation method. Allowed values are: nearest, linear, slinear, cubic.
                For high-dimensional data with large numbers of nodes (> 1E6), it is recommended to 
                use one of the two simpler interpolation methods (nearest or linear) to reduce memory 
                usage.
                For data that require interpolation in latitude and longitude only (e.g. bathymetry) 
                the default interpolation method is a bivariate spline method adapted specifically to 
                spherical coordinates with proper handling of the poles and anti-meridian discontinuity.
            name: str
                Name used to identify the interpolator. Optional.
            max_size: int
                Maximum size (no. bins) along any of the axes of the regular interpolation grid. 
                Optional. Only relevant if input data is on an irregular grid.
            grid_shape: dict()
                Shape of the regular interpolation grid. Optional. Only relevant if input data is on an irregular grid.
            bin_size: dict()
                Bin sizes of the regular interpolation grid. Optional. Overwrites `grid_shape`.
                Only relevant if input data is on an irregular grid.
            irreg_method: str
                Interpolation method used for mapping data from an irregular to a regular grid. 
                Options are `nearest` and `linear`. Default is `nearest`.

        Returns:    
            interp: RegularGridGeospatialInterpolator
                Interpolator
    """

    logger = logging.getLogger("kadlu")

    # convert coordinates to numpy arrays
    coors, _ = _coordinates_as_arrays(kwargs)

    kwargs.update(coors)

    # we will only interpolate dimensions with 2 or more points
    dims, _ = _dim_sizes(coors, min_size=2)

    # if data are on a regular grid, return an instance of the RegularGridGeospatialInterpolator    
    if np.ndim(value) > 1 or len(dims) <= 1:
        return RegularGridGeospatialInterpolator(value, **kwargs)

    # if data are on an irregular grid, map them to a regular grid first using the IrregularGridGeospatialInterpolator class
    reg_grid, coverage = _create_regular_grid(return_coverage=True, **kwargs)

    # debug 
    dims, dim_sizes = _dim_sizes(reg_grid)
    debub_msg = f"Created regular grid in {dims} with shape {dim_sizes} and {coverage*100:.1f}% coverage"
    logger.debug(debub_msg)

    # if coverage is low, issue a warning
    if coverage < 0.5:
        warn_msg = f"The data points only cover {coverage*100:.1f}% of the full rectangular box"
        logger.warning(warn_msg)

    # method and name for interpolator on regular grid
    method = kwargs.pop("method", None)
    name = kwargs.pop("name", None)

    # method and name for interpolator on irregular grid
    irreg_method = kwargs.pop("irreg_method", "nearest")
    irreg_name = "irreg"
    if name is not None:
        irreg_name += "_" + name

    # map from irregular -> regular grid using 'linear' interpolation
    irreg_interp = IrregularGridGeospatialInterpolator(value, method=irreg_method, name=irreg_name, **kwargs)
    v = irreg_interp(**reg_grid, grid=True)

    logger.debug(f"Mapped {len(v) if np.ndim(v) > 0 else 1} values to regular grid using method `{irreg_interp.method}`")

    # return an instance of the RegularGridGeospatialInterpolator on the new, regular grid
    return RegularGridGeospatialInterpolator(value=v, method=method, **reg_grid, name=name)


def _dim_sizes(coors, min_size=1):
    """
    
        TODO: function returns a single dict() argument with key=dim and value=dim_size
    """

    dims, dim_sizes = [], []
    for dim in GEOSPATIAL_DIMS:
        if dim in coors:
            siz = len(np.unique(coors[dim]))
            if siz >= min_size:
                dims.append(dim)
                dim_sizes.append(siz)

    return dims, dim_sizes


def _derivative_orders(kwargs):
    """ Helper function for extracting derivative orders dlat, dlon, ddepth, depoch
        from a set of keyword arguments.
    """
    deriv_orders = {dim: 0 for dim in GEOSPATIAL_DIMS}
    for k,v in kwargs.items(): 
        if k[0] == "d" and k[1:] in GEOSPATIAL_DIMS:
            deriv_orders[k[1:]] = 0 if v is None else v

    return deriv_orders


def _coordinates_as_arrays(kwargs, dtype=np.float64):
    """ Helper function for extracting lat,lon,depth,epoch coordinates 
        from a set of keyword arguments and recasting them as numpy arrays.

        Args:
            kwargs: dict
                The dictionary containing the coordinate arrays
            dtype: 
                The desired data type of the numpy arrays. Default is np.float64

        Returns:
            arrays: dict
                The coordinate arrays, recasted as numpy arrays
            scalar: bool
                True if all cordinates had scalar values, and False otherwise.
    """
    arrays = dict()
    scalar = True
    for k in GEOSPATIAL_DIMS:
        v = kwargs.get(k, None)
        if v is not None:
            scalar *= (np.ndim(v) == 0)
            arrays[k] = as_array(v, dtype=dtype)

    return arrays, scalar


def _assert_same_size(*ai):
    """ Helper function for asserting that a set of arrays have the same size"""
    for a in ai:
        assert a.shape == ai[0].shape, "all arrays must have the same size"


def _reshape(a, dims, coors, grid, scalar):
    """ Helper function for reshaping the interpolation output.

        Args:
            a: numpy array
                The input array
            dims: list(str)
                The names of the dimensions of the input array
            coors: dict
                The coordinate arrays
            grid: bool
                How to combine coordinate elements. If False (default) the coordinate arrays must have matching lengths.
            scalar: bool
                True if all coordinates were specified as scalar values, False otherwise.

        Returns:
            a: numpy array
                The reshaped array
    """
    # make a copy so we don't modify the input argument
    dims = dims.copy()

    # dimensionless input
    if dims is None or len(dims) == 0:
        assert np.ndim(a) == 0 or a.shape == (1,), "input array and `dims` have incompatible shapes"
        
        if coors is None or len(coors) == 0:
            return np.squeeze(a)

        # output shape should match coordinate arrays
        if grid:
            shape = np.meshgrid(*coors.values(), indexing="ij")[0].shape
        else:
            shape = len(list(coors.values())[0])

        a = np.ones(shape) * np.squeeze(a)

        if scalar:
            a = np.squeeze(a)

        return a

    # array input
    assert len(coors) > 0, "at least one coordinate array required for reshaping"

    if not grid:
        _assert_same_size(a, *coors.values())
        if scalar:
            a = float(np.squeeze(a))

        return a

    assert len(a.shape) == len(dims), f"input array and `dims` have incompatible shapes: a.shape={a.shape}, len(dims)={len(dims)}"

    # expand the array by adding new dimensions
    for k,v in coors.items():
        if k not in dims:
            new_shape = a.shape + (len(v),)
            a = np.expand_dims(a, axis=-1)  #adds new dimension of shape (1,)
            a = np.broadcast_to(a, new_shape)  #expands the new dimension to the desired size (by replicating the array)
            dims.append(k)

    # re-order the axes to match the specified ordering
    indices = [dims.index(d) for d in GEOSPATIAL_DIMS if d in dims]
    a = np.moveaxis(a, indices, np.arange(len(indices)))

    # always return a (Python) scalar value for scalar coordinates
    if scalar:
        a = float(np.squeeze(a))

    return a


def _complete_grid(**kwargs):
    """ Helper function for adding depth and time dimensions to a lat-lon grid
    
        Args:
            lat: 2d numpy array
                The latitude coordinates of the lat-lon grid
            lon: 2d numpy array
                The longitude coordinates of the lat-lon grid
            depth: array-like
                The depth coordinates, in meters
            epoch: array-like
                The time coordinates, in epoch hours since 2000-01-01 00:00:00
        
        Returns:
            flattened_coors: dict
                The flattened coordinate arrays
            grid_shape: tuple
                The (lat,lon,depth,epoch) grid shape    
    """
    args, _ = _coordinates_as_arrays(kwargs)

    coors = {
        "lat": args.pop("lat"),
        "lon": args.pop("lon"),
    }

    # loop ove new dimensions (depth, epoch)
    for new_dim in args.keys():
        u = args[new_dim]

        for dim in coors.keys():
            # expand existing dimensions
            v = coors[dim]
            new_shape = v.shape + (len(u),)
            v = np.expand_dims(v, axis=-1) 
            coors[dim] = np.broadcast_to(v, new_shape)  

        # expand the new dimension            
        axis = tuple(np.arange(len(coors["lat"].shape)))
        coors[new_dim] = np.ones(coors["lat"].shape) * np.expand_dims(u, axis=axis)

    # the full grid shape
    grid_shape = coors["lat"].shape

    # flattened coordinate arrays
    flattened_coors = {k: v.flatten() for k,v in coors.items()}

    return flattened_coors, grid_shape


class BaseGeospatialInterpolator():
    """ Parent class for all interpolators.

        Child classes must implement the `_eval` method.        

        Interpolators accept the following coordinates,

            * lat: latitude in degrees
            * lon: longitudes in degrees
            * depth: depth in meters below the sea surface
            * epoch: time in hours since 2000-01-01 00:00:00
    
        Args:
            value: array-like
                Data values
            lat: array-like
                Latitudes in degrees. 
            lon: array-like
                Longitudes in degrees. 
            depth: array-like
                Depths in meters below the sea surface. 
            epoch: array-like
                Time in hours since 2000-01-01 00:00:00. 
            name: str
                Name used to identify the interpolator. 
            origin: tuple(float,float)
                Reference location used as the origo of XY coordinate system. If not specified, 
                the center point of the lat,lon coordinates will be used.

        Attrs:
            value: numpy array
                Data values
            coordinates: dict
                Coordinates arrays
    """
    def __init__(
        self, 
        value,
        name="GeospatialInterpolator",
        origin=None,
        **coors,
    ):
        self.name = name
        self.logger = logging.getLogger("kadlu")

        # convert values and coordinates to numpy arrays
        self.value = as_array(value)
        self.coordinates, _ = _coordinates_as_arrays(coors)

        # lat,lon reference location for XY coordinate system
        self.origin = origin

        if self.origin is None:
            self.origin = center_point(self.coordinates.get("lat"), self.coordinates.get("lon"))

        if self.origin is None:
            self.origin = (0, 0)

    def __call__(self, **kwargs):
        """ Evaluate the interpolation at the given coordinate(s)

            Location coordinates can be specified either as (lat,lon) or as 
            (x,y) where x and y refer to the E-W and N-S axes of a planar 
            coordinate system centered at the `origin` reference location.

            Args:
                lat: array-like
                    Latitudes in degrees.
                lon: array-like
                    Longitudes in degrees.
                depth: array-like
                    Depths in meters below the sea surface. 
                epoch: array-like
                    Time in hours since 2000-01-01 00:00:00. 
                x: array-like
                   x coordinates in meters. 
                y: array-like
                   y coordinate in meters. 
                grid: bool
                    How to combine coordinate elements. If False (default) the coordinate arrays must have matching lengths.      
                origin: tuple(float,float)
                    Reference location for XY coordinate system. If not specified, the reference location specified at 
                    initialisation is used. Only relevant if location coordinates are specified as x,y instead of lat,lon.
                dlat: int
                    Order of derivative in latitude.
                dlon: int
                    Order of derivative in longitude.
                ddepth: int
                    Order of derivative in depth.
                depoch: int
                    Order of derivative in time.
                dx: int
                    Order of derivative in x.
                dy: int
                    Order of derivative in y.
        """
        if "x" in kwargs or "y" in kwargs: 
            return self._eval_xy(**kwargs)
    
        else:
            return self._eval(**kwargs)

    def _eval(self, **kwargs):
        """ Must be implemented in child class """
        pass

    def _eval_xy(self, **kwargs):
        kwargs = kwargs.copy()

        grid = kwargs.pop("grid", False)
        origin = kwargs.pop("origin", self.origin)

        # map x,y to lat,lon
        lat, lon = XYtoLL(
            x=kwargs.pop("x", 0),
            y=kwargs.pop("y", 0),
            lat_ref=origin[0],
            lon_ref=origin[1],
            grid=grid
        )

        kwargs.update({"lat": lat, "lon": lon})

        # complete lat-lon grid by including depth and time dimensions
        if grid:
            flattened_coors, grid_shape = _complete_grid(**kwargs)
            kwargs.update(flattened_coors)

        # pass flattened coordinate arrays to lat-lon interpolation method
        v = self._eval(grid=False, **kwargs)

        # derivatives
        kwargs["dlat"] = kwargs.get("dy", 0)
        kwargs["dlon"] = kwargs.get("dx", 0)

        # for derivates, we must multiply by the appropriate Jacobian
        d = _derivative_orders(kwargs)
        if d["lat"] + d["lon"] > 0:
            # convert from deg^-1 to rad^-1
            v /= np.power(deg2rad, d["lat"] + d["lon"])

            # multiply by jacobian
            j = DLDL_over_DXDY(
                lat=lat,
                lat_deriv_order=d["lat"],
                lon_deriv_order=d["lon"],
            )
            v *= j

        # reshape output to desired shape
        if grid:
            v = np.reshape(v, newshape=grid_shape)
            v = np.swapaxes(v, 0, 1)

        return v


class RegularGridGeospatialInterpolator(BaseGeospatialInterpolator):
    """

        Args:
            method: str
                For interpolation of data on a 2d lat-lon grid, it is recommended to 
                leave `method` unspecified (i.e. method=None)

    """
    def __init__(
        self, 
        value,
        name="RegularGridGeospatialInterpolator",
        origin=None,
        method=None,
        **coors,
    ):
        super().__init__(value=value, name=name, origin=origin, **coors)

        # deduce the shape of the interpolation grid from the sizes of the coordinate arrays
        grid_shape = tuple([
            len(self.coordinates[dim]) for dim in GEOSPATIAL_DIMS if dim in self.coordinates
        ])

        # we will only interpolate dimensions with 2 or more points
        self.dims, self.dim_sizes = _dim_sizes(self.coordinates, min_size=2)

        # validate array shapes
        assert_msg = f"value and coordinate arrays have incompatible shapes,  value:{self.value.shape}, coordinates:{grid_shape}"
        assert np.ndim(value) == 0 or self.value.shape == grid_shape or self.value.shape == tuple(self.dim_sizes), assert_msg

        debug_msg = f"[{self.name}] Detected {len(self.dims)} dimensions that require interpolation: {self.dims}"
        self.logger.debug(debug_msg)

        # grab dimensions that require interpolation
        points = [v for k,v in self.coordinates.items() if k in self.dims]
        
        # drop dimensions that don't require interpolation
        values = np.squeeze(self.value)

        # if lat,lon are the only dimensions requiring interpolation, and each has at least 3 points, we can use scipy's RectSphereBivariateSpline
        if self.dims == ["lat", "lon"] and np.all(np.array(self.dim_sizes) >= 3) and method == None:
            self._interp = _RectSphereBivariateSpline(
                points=[self.coordinates["lat"], self.coordinates["lon"]],
                values=values
            )
        
        # in all other cases, we use scipy's RegularGridInterpolator
        elif len(self.dims) > 0:
            self._interp = _RegularGridInterpolator(
                points=points,
                values=values,
                method="linear" if method is None else method,
                bounds_error=False,
                fill_value=None,  #values outside the interpolation domain will be extrapolated
            )

        # except in the trivial case where no dimensions require interpolation
        else:
            self._interp = _ConstantInterpolator(values)

    @property
    def method(self):
        return self._interp.method

    def _eval(self, grid=False, **kwargs):
        # coordinates where interpolation is to be evaluated
        coors, scalar = _coordinates_as_arrays(kwargs)

        # check that the coordinates required for evaluating the interpolation were provided
        for dim in self.dims:
            assert dim in coors, f"[{self.name}] `{dim}` required for evaluating interpolation"

        # check that coordinate arrays have same size
        # TODO: move this check up to the BaseGeospatialInterpolator ?
        if not grid:
            _assert_same_size(*coors.values())

        # grab dimensions that require interpolation
        xi = [v for k,v in coors.items() if k in self.dims]

        # pass to interpolator
        v = self._interp(xi, grid=grid, **kwargs)

        # reshape output to desired shape
        v = _reshape(v, dims=self.dims, coors=coors, grid=grid, scalar=scalar)

        return v


class IrregularGridGeospatialInterpolator(BaseGeospatialInterpolator):
    """
    
        Attrs:
            dims: list(str)
                Names of the dimensions that are being interpolated
    """
    def __init__(
        self, 
        value,
        name="IrregularGridGeospatialInterpolator",
        origin=None,
        method="cubic", 
        **coors,
    ):
        super().__init__(value=value, name=name, origin=origin, **coors)

        # coordinate arrays must have the same size
        _assert_same_size(*self.coordinates.values())

        # we will only interpolate dimensions with 2 or more unique values
        self.dims, _ = _dim_sizes(self.coordinates, min_size=2)
        debug_msg = f"[{self.name}] Detected {len(self.dims)} dimensions that require interpolation: {self.dims}"
        self.logger.debug(debug_msg)

        # stack that coordinates that require interpolation as an array with shape (num_pts, num_dims)
        if len(self.dims) > 0:
            points = np.stack([v for k,v in self.coordinates.items() if k in self.dims], axis=-1)
            self._interp = _IrregularGridInterpolator(points, self.value, method=method)

        else:
            self._interp = _ConstantInterpolator(self.value)

    @property
    def method(self):
        return self._interp.method

    def _eval(self, grid=False, replace_nan=True, method=None, **kwargs):
        # coordinates where interpolation is to be evaluated
        coors, scalar = _coordinates_as_arrays(kwargs)

        # check that the coordinates required for evaluating the interpolation were provided
        for dim in self.dims:
            assert dim in coors, f"{self.name} `{dim}` required for evaluating interpolation"

        # check that coordinate arrays have same size
        # TODO: move this check up to the BaseGeospatialInterpolator ?
        if not grid:
            _assert_same_size(*coors.values())
                
        # grab dimensions that require interpolation
        xi = [v for k,v in coors.items() if k in self.dims]

        # pass to interpolator
        v = self._interp(xi, grid=grid, replace_nan=replace_nan, method=method, **kwargs)

        # reshape output to desired shape
        v = _reshape(v, dims=self.dims, coors=coors, grid=grid, scalar=scalar)
        
        return v


class _RegularGridInterpolator(RegularGridInterpolator):
    """ Helper class for RegularGridGeospatialInterpolator
    
    """

    def __init__(self, points, values, **kwargs):
        self.name = self.__class__.__name__
        self.logger = logging.getLogger("kadlu")

        # validate method
        method = kwargs.pop("method", "slinear")
        self.method = self._validate_method(method, points)

        # initialize interpolator
        super().__init__(
            points=points,
            values=values,
            method=self.method,
            **kwargs
        )

    def _validate_method(self, method, points):
        """ Helper function for validating the requested interpolation method
        
            Args:
                method: str
                    Interpolation method.
                points: tuple(array)
                    Coordinate arrays

            Returns:
                method: str
                    Same as input if validation was successful; otherwise returns an 
                    alternative method compatible with the size of the coordinate arrays

            Raises:
                AssertionError: if the requested method is invalid
        """ 
        # minimum no. data points required for each of the interpolation methods offered by scipy's RegularGridInterpolator
        self.min_data_pts = {
            "nearest": 1,
            "slinear": 2,
            "linear": 2,
            "cubic": 4,
            #"quintic": 6,
        }

        assert_msg = f"[{self.name}] Invalid interpolation method `{method}`. Valid options are: {list(self.min_data_pts.keys())}."
        assert method in self.min_data_pts, assert_msg

        # check that we have sufficient points for the chosen interpolation method
        # automatically switch to an alternative method that meets the requirement, if needed
        min_siz = np.min([len(arr) for arr in points])
        if min_siz < self.min_data_pts[method]:
            sorted_dict = sorted(self.min_data_pts.items(), key=lambda item: item[1], reverse=True)
            for k,v in sorted_dict:
                if v <= min_siz:
                    alt_method = k 
                    break          

            warn_msg = f"[{self.name}] Too few points for `{method}` interpolation. Switching to `{alt_method}`"
            self.logger.warning(warn_msg)
            method = alt_method

        return method

    def __call__(self, xi, grid=False, **kwargs):
        n_deriv = np.sum([n for n in _derivative_orders(kwargs).values()])
        if n_deriv > 0:
            err_msg = f"[{self.name}] Interpolation of derivatives only implemented for spherical (lat,lon) geometries"
            raise NotImplementedError(err_msg)

        if grid and len(xi) > 1:
            xi = np.meshgrid(*xi, indexing="ij")
            grid_shape = xi[0].shape
            xi = [x.flatten() for x in xi]

        pts = np.array(xi).T
        v = super().__call__(pts)

        if grid and len(xi) > 1: 
            # reshape flattened array back to grid shape
            v = np.reshape(v, newshape=grid_shape)

        return v


class _RectSphereBivariateSpline(RectSphereBivariateSpline):
    """ Helper class for RegularGridGeospatialInterpolator
    
    """

    def __init__(self, points, values, **kwargs):
        u, v = torad(points[0], points[1])

        super().__init__(
            u=u,
            v=v,
            r=values,
            **kwargs
        )

        self.method = "spline"

    def __call__(self, xi, grid=False, dlat=0, dlon=0, **kwargs):
        lats_rad, lons_rad = torad(xi[0], xi[1])

        # when `grid` is True, we must ensure latitudes and longitudes are strictly increasing
        if grid:
            lat_indices = np.argsort(lats_rad)
            lats_rad = lats_rad[lat_indices]
            lon_indices = np.argsort(lons_rad)
            lons_rad = lons_rad[lon_indices]

        # evaluate the interpolation function
        v = super().__call__(
            theta=lats_rad,
            phi=lons_rad,
            grid=grid,
            dtheta=dlat,
            dphi=dlon
        )

        # convert derivatives from rad^-1 to deg^-1
        if dlat + dlon > 0:
            v *= np.power(deg2rad, dlat + dlon)

        if grid:
            # reverse the index mapping so lat,lon values appear in their original order
            indices = reverse_index_map(lat_indices)
            v = v[indices]
            v = np.swapaxes(v, 0, 1)
            indices = reverse_index_map(lon_indices)
            v = v[indices]
            v = np.swapaxes(v, 0, 1)

        return v


class _IrregularGridInterpolator():
    """ Helper class for IrregularGridGeospatialInterpolator
    
    """

    def __init__(
        self, 
        points,
        values, 
        method="cubic", 
        fill_value=np.nan,
        batch_size=10000,
        **kwargs,
    ):
        self.name = self.__class__.__name__
        self.logger = logging.getLogger("kadlu")

        self.points = points
        self.values = values

        # available interpolation methods, in order of preference
        self.methods = ["cubic", "linear", "nearest"]
        
        # `cubic` only available for 1D and 2D data
        if self.points.shape[1] > 2: 
            del self.methods[0]
            if method == "cubic":
                method = "linear"

        self._validate_method(method)

        self.method = method
        self.fill_value = fill_value
        self.batch_size = batch_size

    def _validate_method(self, method):
        """ Helper function for validating the requested interpolation method
        
            Args:
                method: str
                    Interpolation method.

            Raises:
                AssertionError: if the requested method is invalid
        """ 
        assert_msg = f"[{self.name}] Invalid interpolation method `{method}`. Valid options are: {self.methods}."
        assert method in self.methods, assert_msg

    def __call__(self, xi, grid=False, replace_nan=True, method=None, **kwargs):
        n_deriv = np.sum([n for n in _derivative_orders(kwargs).values()])
        if n_deriv > 0:
            err_msg = f"[{self.name}] Interpolation of derivatives not implemented for irregular grids"
            raise NotImplementedError(err_msg)

        if method is None:
            method = self.method

        self._validate_method(method)

        if grid and len(xi) > 1:
            xi = np.meshgrid(*xi, indexing="ij")
            grid_shape = xi[0].shape
            xi = [x.flatten() for x in xi]

        pts = np.array(xi).T

        # alternative interpolation methods to try, if the selected one fails
        alt_methods = self.methods.copy()
        alt_methods.remove(method)

        counter = 0 #attempt counter
        while True:
            try:
                # evaluate the interpolation function
                # (split into smaller batches to not use excessive memory)

                debug_msg = f"[{self.name}] Interpolating {pts.shape[0]} data points on {pts.shape[1]}D irregular grid"\
                    + f" using method `{method}` and batch size of {self.batch_size}"
                self.logger.debug(debug_msg)

                v = self._batched_eval(pts, method)

                break

            except (Exception, QhullError):
                msg = f"[{self.name}] Interpolation on irregular grid with method=`{method}` failed. To view full error report, re-run in DEBUG mode." 
                self.logger.debug(traceback.format_exc())

                # try again, with a different interpolation method; continue until successful or run out of options
                if counter < len(alt_methods):
                    method = alt_methods[counter]
                    counter += 1

                    msg += f" Switching to method=`{method}`."
                    self.logger.warning(msg)

                else:
                    raise Exception(msg)
                
        self.method = method

        # replace NaN values with nearest valid value
        if replace_nan and method != "nearest":
            indices_nan = np.where(np.isnan(v))
            nearest_value = self._batched_eval(pts, "nearest")
            v[indices_nan] = nearest_value[indices_nan]

        # if griddata return a 2d array with same (num_points, 1) we drop the last axis
        if np.ndim(v) == 2:
            v = v[:, 0]

        if grid and len(xi) > 1: 
            # reshape flattened array back to grid shape
            v = np.reshape(v, newshape=grid_shape)

        return v

    def _batched_eval(self, pts, method):
        """ Helper function for splitting evaluation of griddata function into multiple smaller requests"""
        n_batches = int(np.ceil(pts.shape[0] / self.batch_size))
        v = []
        if n_batches > 1:
            info_msg = f"Interpolating {pts.shape[0]} data points on {pts.shape[1]}D irregular grid"\
                + f" using method `{method}` and batch size of {self.batch_size}"
            self.logger.info(info_msg)

        for i in tqdm(range(n_batches), disable = n_batches < 2):
            a = i * self.batch_size
            b = a + self.batch_size
            vi = griddata(self.points, self.values, pts[a:b], method=method, fill_value=self.fill_value, rescale=False)
            v.append(vi)

        v = np.concatenate(v, axis=0)
        return v


class _ConstantInterpolator():
    """ Convenience class for handling constant/uniform data.
    """

    def __init__(self, value):
        self.value = np.squeeze(value)
        self.method = None

    def __call__(self, *args, **kwargs):
        n_deriv = np.sum([n for n in _derivative_orders(kwargs).values()])
        if n_deriv > 0:
            return 0
        else:
            return self.value


def _create_regular_grid(max_size=100, grid_shape=None, bin_size=None, return_coverage=False, **kwargs):
    """ Creates regular grid with uniform spacing that covers a set of (lat,lon,depth,epoch) coordinates.

        If grid_shape=None and bin_size=None, the grid spacing is set to the smallest non-zero difference 
        between two coordinate values. If necessary, the grid spacing is inflated so as to not exceed the 
        specified max size.

        Args:
            max_size: int
                Maximum size (no. bins) along any of the axes of the grid. Optional. Default is 100.
            grid_shape: dict()
                Grid shape. Optional.
            bin_size: dict()
                Bin sizes. Optional. Overwrites `grid_shape`.
            return_coverage: bool
                Also compute the grid's *coverage*. The coverage is a number between 0-1 that indicates how 
                well the input coordinates cover the volume spanned by the regular grid. It is computed as 
                the  ratio of the convex hull the input coordinates to the volume of the created grid. 
                
        Keyword args:                
            lat: array-like
                Latitudes in degrees. 
            lon: array-like
                Longitudes in degrees. 
            depth: array-like
                Depths in meters below the sea surface. 
            epoch: array-like
                Time in hours since 2000-01-01 00:00:00. 

        Returns:
            reg_coors: dict
                The coordinates of the regular grid
            coverage: float
                Only returned if `return_coverage=True`
    """
    # coordinate arrays, ordered as lat,lon,depth,epoch
    coors, _ = _coordinates_as_arrays(kwargs)

    # coordinate min/max values
    coors_range = {k: [np.min(v), np.max(v)] for k,v in coors.items()}

    # if the grid shape is not specified, use the smallest non-zero coordinate 
    # spacing as bin size, inflated if necessary to not exceed the max size
    if grid_shape is None:
        grid_shape = dict()

        for k,arr in coors.items():
            unique_values = np.unique(arr)

            if len(unique_values) == 1:
                n_nodes = 1

            else:
                start, stop = coors_range[k][0], coors_range[k][1]
                if bin_size is None:
                    step = np.min(np.diff(np.sort(unique_values)))
                else:
                    step = bin_size[k]
                
                n_nodes = int((stop - start) / step) + 1

            if max_size is not None:
                n_nodes = min(max_size, n_nodes)

            grid_shape[k] = n_nodes 

    # regularly spaced coordinates
    reg_coors = dict()
    for k,arr in coors.items():
        start, stop = coors_range[k][0], coors_range[k][1]
        num = grid_shape[k]
        reg_coors[k] = np.unique(np.linspace(start=start, stop=stop, num=num))

    if not return_coverage:
        return reg_coors

    # compute "coverage"
    
    # only consider dimensions with at least two unique values
    dims, _ = _dim_sizes(reg_coors, min_size=2)

    # volume of the regular grid
    grid_vol = 1
    for k in dims:
        start, stop = coors_range[k][0], coors_range[k][1]
        grid_vol *= (stop - start)

    # convex hull of the input coordinates
    pts = np.array([arr for k,arr in coors.items() if k in dims])
    pts = np.swapaxes(pts, 0, 1)
    h = ConvexHull(pts)

    coverage = h.volume / grid_vol

    return reg_coors, coverage        


""" Unit tests for the the 'sound.geophony' module in the 'kadlu' package

    Authors: Oliver Kirsebom
    contact: oliver.kirsebom@dal.ca
    Organization: MERIDIAN-Intitute for Big Data Analytics
    Team: Acoustic data Analytics, Dalhousie University
    Project: packages/kadlu
             Project goal: Tools for underwater soundscape modeling

    License:

"""

import pytest
import os
import numpy as np
from kadlu.sound.geophony import geophony, transmission_loss, kewley_sl_func, source_level
from kadlu.geospatial.ocean import Ocean
from kadlu.utils import R1_IUGG, deg2rad

current_dir = os.path.dirname(os.path.realpath(__file__))
path_to_assets = os.path.join(os.path.dirname(current_dir), "assets")


def test_kewley_sl_func():
    sl1 = kewley_sl_func(freq=10, wind_uv=0)
    sl2 = kewley_sl_func(freq=40, wind_uv=2.57)
    assert sl1 == sl2
    assert sl2 == 40.0
    sl3 = kewley_sl_func(freq=40, wind_uv=5.14)
    assert sl3 == 44.0
    sl4 = kewley_sl_func(freq=100, wind_uv=5.14)
    assert sl4 == 42.5


def test_source_level():
    ok = {'load_bathymetry': 10000, 'load_wind_uv': 5.14}
    o = Ocean(**ok)
    sl = source_level(freq=10,
                      x=0,
                      y=0,
                      area=1,
                      ocean=o,
                      sl_func=kewley_sl_func)
    assert sl == 44.0
    sl = source_level(freq=100,
                      x=[0, 100],
                      y=[0, 100],
                      area=[1, 2],
                      ocean=o,
                      sl_func=kewley_sl_func)
    assert sl[0] == 42.5
    assert sl[1] == sl[0] + 10 * np.log10(2)


def test_geophony_flat_seafloor():
    """ Check that we can execute the geophony method for a
        flat seafloor and uniform sound speed profile"""
    kwargs = {
        'load_bathymetry': 10000,
        'load_wind_uv': 1.0,
        'ssp': 1480,
        'angular_bin': 90,
        'dr': 1000,
        'dz': 1000
    }
    geo = geophony(freq=100,
                   south=44,
                   north=46,
                   west=-60,
                   east=-58,
                   depth=[100, 2000],
                   xy_res=71,
                   **kwargs)
    spl = geo['spl']
    x = geo['x']
    y = geo['y']
    assert x.shape[0] == 3
    assert y.shape[0] == 5
    assert spl.shape[0] == 3
    assert spl.shape[1] == 5
    assert spl.shape[2] == 2
    assert np.all(np.diff(x) == 71e3)
    assert np.all(np.diff(y) == 71e3)
    # try again, but this time for specific location
    kwargs = {
        'load_bathymetry': 10000,
        'load_wind_uv': 1.0,
        'ssp': 1480,
        'angular_bin': 90,
        'dr': 1000,
        'dz': 1000,
        'propagation_range': 50
    }
    geo = geophony(freq=100, lat=45, lon=-59, depth=[100, 2000], **kwargs)


def test_geophony_in_canyon(bathy_canyon):
    """ Check that we can execute the geophony method for a
        canyon-shaped bathymetry and uniform sound speed profile"""
    kwargs = {
        'load_bathymetry': bathy_canyon,
        'load_wind_uv': 1.0,
        'ssp': 1480,
        'angular_bin': 90,
        'dr': 1000,
        'dz': 1000
    }
    z = [100, 1500, 3000]
    geo = geophony(freq=10,
                   south=43,
                   north=46,
                   west=60,
                   east=62,
                   depth=z,
                   xy_res=71,
                   **kwargs)
    spl = geo['spl']
    x = geo['x']
    y = geo['y']
    assert spl.shape[0] == x.shape[0]
    assert spl.shape[1] == y.shape[0]
    assert spl.shape[2] == len(z)
    assert np.all(np.diff(x) == 71e3)
    assert np.all(np.diff(y) == 71e3)
    # check that noise is NaN below seafloor and non Nan above
    bathy = np.swapaxes(
        np.reshape(geo['bathy'], newshape=(y.shape[0], x.shape[0])), 0, 1)
    bathy = bathy[:, :, np.newaxis]
    xyz = np.ones(shape=bathy.shape) * z
    idx = np.nonzero(xyz >= bathy)
    assert np.all(np.isnan(spl[idx]))
    idx = np.nonzero(xyz < bathy)
    assert np.all(~np.isnan(spl[idx]))


@pytest.mark.hycom_access
def test_transmission_loss_real_world_env():
    """ Check that we can initialize a transmission loss object
        for a real-world environment and obtain the expected result 
        
        #TODO: understand why test fails for 'cubic' interpolation
                has to do with poor interpolation of temperature and salinity for certain parts of the coordinate space ...
        
    """

    from datetime import datetime
    bounds = dict(start=datetime(2019, 1, 1),
                  end=datetime(2019, 1, 2),
                  top=0,
                  bottom=10000)
    src = dict(load_bathymetry='gebco',
               load_temperature='hycom',
               load_salinity='hycom')
    sound_source = {
        'freq': 200,
        'lat': 43.8,
        'lon': -59.04,
        'source_depth': 12
    }
    seafloor = {'sound_speed': 1700, 'density': 1.5, 'attenuation': 0.5}
    transm_loss, ocean = transmission_loss(
        seafloor=seafloor,
        propagation_range=20,
        **src,
        **bounds,
        **sound_source,
        dr=100,
        angular_bin=45,
        dz=50,
        #drop="epoch", #drop time dimension
        #interp_args={"method":"linear"},  
        return_ocean=True,
    )

    # check fetching and interpolation of ocean variables
    lats = np.linspace(43.7, 43.9, num=3)
    lons = np.linspace(-59.14, -58.94, num=3)
    t = datetime(2019, 1, 1, 12) 
    t0 = datetime(2000, 1, 1)
    epoch = (t - t0).total_seconds() / 3600 #hours since since 2000-01-01 00:00
    seafloor_depth = ocean.bathymetry(lat=lats, lon=lons, grid=True)
    max_depth = 1712.83265634
    depths = np.linspace(0, max_depth, num=3)
    temp = ocean.temperature(lat=lats, lon=lons, depth=depths, epoch=epoch, grid=True)
    salinity = ocean.salinity(lat=lats, lon=lons, depth=depths, epoch=epoch, grid=True)

    temp = temp[:,:,0,:]
    salinity = salinity[:,:,0,:]

    answ_seafloor_depth = np.array([[266.47, 606.66, 1474.88], [101.38, 270.35, 1632.57], [84.09, 398.28, 1407.04]])
    answ_temp = np.array([[[3.23, 4.24, 3.78], [3.67, 4.15, 3.78], [4.25, 4.23, 3.79]], [[2.84, 4.18, 3.78], [3.02, 4.18, 3.79], [3.46, 4.2, 3.79]], [[2.88, 4.16, 3.79], [2.85, 4.16, 3.79], [3.01, 4.22, 3.76]]])
    answ_salinity = np.array([[[32.4, 34.93, 34.95], [32.48, 34.93, 34.95], [32.79, 34.93, 34.95]], [[32.26, 34.93, 34.95], [32.36, 34.93, 34.95], [32.49, 34.93, 34.95]], [[31.92, 34.93, 34.95], [32.17, 34.93, 34.95], [32.29, 34.93, 34.96]]])

    #print(seafloor_depth.round(2).tolist())
    #print(temp.round(2).tolist())
    #print(salinity.round(2).tolist())

    np.testing.assert_array_almost_equal(seafloor_depth, answ_seafloor_depth, decimal=2)
    np.testing.assert_array_almost_equal(temp, answ_temp, decimal=2)
    np.testing.assert_array_almost_equal(salinity, answ_salinity, decimal=2)

    # check transmission loss calculation results
    tl_h, ax_h, tl_v, ax_v = transm_loss.calc(vertical=True)

    answ_h = np.array(
        [[97.5, 116.9, 116.9, 117.0, 120.7, 120.1, 121.4, 122.4, 123.5, 124.4], [97.5, 114.9, 113.9, 119.4, 122.4, 120.6, 120.6, 120.4, 120.2, 123.2], [97.5, 114.2, 116.6, 118.7, 119.9, 119.6, 121.9, 121.6, 120.9, 124.0], [97.5, 113.9, 118.7, 119.3, 116.4, 120.4, 123.3, 119.1, 120.7, 126.3], [97.5, 115.0, 121.2, 118.9, 120.1, 119.3, 120.1, 122.1, 122.2, 122.5], [97.5, 116.8, 115.3, 117.1, 123.4, 118.9, 118.4, 123.7, 121.9, 120.5], [97.5, 115.4, 119.6, 118.6, 118.4, 120.3, 123.6, 119.4, 123.4, 121.8], [97.5, 116.5, 119.7, 118.6, 120.0, 121.3, 122.5, 123.6, 124.5, 125.5]]
    )

    answ_v = np.array(
        [[31.9, 65.4, 68.5, 70.0, 71.1, 72.1, 73.4, 74.5, 75.6, 76.5, 77.3], [53.4, 142.2, 161.3, 144.4, 141.4, 143.5, 144.7, 145.5, 146.6, 147.5, 148.4], [59.5, 158.7, 176.4, 161.4, 158.5, 160.5, 161.7, 162.5, 163.6, 164.5, 165.4], [63.7, 168.8, 186.2, 171.7, 168.7, 170.7, 172.0, 172.8, 173.9, 174.8, 175.7], [67.5, 176.4, 193.7, 179.3, 176.4, 178.4, 179.6, 180.4, 181.5, 182.4, 183.3], [71.6, 183.0, 200.2, 185.9, 183.0, 184.9, 186.2, 187.0, 188.1, 189.0, 189.9], [77.3, 194.4, 211.7, 195.6, 194.7, 196.9, 198.3, 199.0, 200.1, 200.9, 201.7], [93.1, 212.6, 220.4, 213.7, 215.8, 217.6, 219.0, 219.7, 220.8, 221.6, 222.4]]
    )

    res_h = tl_h[0, 0, :, ::20]
    assert np.all(np.abs(res_h - answ_h) < 0.03 * np.abs(answ_h)) #agree within 3%
    #print(np.round(res_h, 1).tolist())

    res_v = tl_v[0, 1::10, ::20, 0]
    assert np.all(np.abs(res_v - answ_v) < 0.06 * np.abs(answ_v)) #agree within 6%
    #print(np.round(res_v, 1).tolist())

    assert tl_h.shape == (1, 1, 8, 200), f'tl_h.shape = {tl_h.shape}'
    assert tl_v.shape == (1, 73, 201, 8), f'tl_v.shape = {tl_v.shape}'


def test_transmission_loss_flat_seafloor():
    """ Check that we can initialize a transmission loss object
        for a flat seafloor and uniform sound speed profile """
    transm_loss = transmission_loss(freq=100,
                                    source_depth=75,
                                    propagation_range=0.5,
                                    load_bathymetry=2000,
                                    ssp=1480,
                                    angular_bin=10)
    tl_h, ax_h, tl_v, ax_v = transm_loss.calc(vertical=True)
    answ = np.genfromtxt(os.path.join(path_to_assets,
                                      'lloyd_mirror_f100Hz_SD75m.csv'),
                         delimiter=",")
    assert answ.shape == tl_v[0, :, :, 0].shape
    np.testing.assert_array_almost_equal(-tl_v[0, 1:, :, 0],
                                         answ[1:, :],
                                         decimal=3)



"""
    Kadlu API for the NOAA WaveWatch III Datastore

    User guides:
        https://github.com/NOAA-EMC/WW3/wiki/WAVEWATCH-III-User-Guide

    Data model description (boundary definitions, map visualizations, etc)
        https://polar.ncep.noaa.gov/waves/implementations.php
"""

import shutil
import logging
import requests
import platform
import warnings
from datetime import datetime, timedelta
from os.path import isfile
from hashlib import md5

import numpy as np
if platform.machine() == 'arm64':
    warnings.warn(
        "WWIII module requires the pygrib package which is not available for this platform"
    )
else:
    import pygrib

from kadlu import index
from kadlu.geospatial.data_sources.data_util import (
    Boundary,
    database_cfg,
    dt_2_epoch,
    logmsg,
    logmsg_nodata,
    storage_cfg,
    str_def,
)

wwiii_tables = ['hs', 'dp', 'tp', 'windU', 'windV']

wwiii_src = "https://data.nodc.noaa.gov/ncep/nww3/"

# region boundaries as defined in WWIII docs:
#    https://polar.ncep.noaa.gov/waves/implementations.php
wwiii_global = Boundary(-90, 90, -180, 180, 'glo_30m')  # global
wwiii_regions = [
    Boundary(15, 47, -99, -60, 'at_4m'),  # atlantic
    Boundary(15, 50, -165, -116, 'wc_4m'),  # US west
    Boundary(48, 74, 140, 180, 'ak_4m'),  # alaska (west)
    Boundary(48, 74, -180, -120, 'ak_4m'),  # alaska (east)
    Boundary(65, 84, -180, 180, 'ao_30m'),  # arctic ocean
    Boundary(-20, 30, 130, 180, 'ep_10m'),  # pacific (west)
    Boundary(-20, 30, -180, -145, 'ep_10m')
]  # pacific (east)


def initdb():
    conn, db = database_cfg()
    for var in wwiii_tables:
        db.execute(f'CREATE TABLE IF NOT EXISTS {var}'
                   '( val     REAL    NOT NULL, '
                   '  lat     REAL    NOT NULL, '
                   '  lon     REAL    NOT NULL, '
                   '  time    INT     NOT NULL, '
                   '  source  TEXT    NOT NULL) ')
        db.execute(f'CREATE UNIQUE INDEX IF NOT EXISTS '
                   f'idx_{var} on {var}(time, lon, lat, val, source)')
    conn.close()


def insert(table, agg):
    """ insert parsed data into local database """
    conn, db = database_cfg()
    db.executemany(
        f"INSERT OR IGNORE INTO {table} VALUES (?,?,?,CAST(? AS INT),?)",
        agg.T)
    conn.commit()
    conn.close()


def fetch_wwiii(var, **kwargs):
    """ download wwiii data and return associated filepaths

        args:
            var: string
                the variable name of desired parameter according to WWIII docs
                the complete list of variables can be found at the following
                URL under 'model output'
                https://polar.ncep.noaa.gov/waves/implementations.php
            south, north: float
                ymin, ymax coordinate boundaries (latitude). range: -90, 90
            west, east: float
                xmin, xmax coordinate boundaries (longitude). range: -180, 180
            start: datetime
                the start of the desired time range
            end: datetime
                the end of the desired time range

        return:
            True if new data was fetched, else False
    """
    assert 6 == sum(
        map(lambda kw: kw in kwargs.keys(),
            ['south', 'north', 'west', 'east', 'start', 'end'
             ])), 'malformed query'
    assert var in ('hs', 'tp', 'dp', 'wind'), 'invalid fetch var'
    assert kwargs['end'].month <= kwargs['start'].month+1 and kwargs['start'].year <= kwargs['end'].year+1, \
            f'fetch function can only load one month at a time. got start: {kwargs["start"]} end: {kwargs["end"]}'

    #print("WWIII NOTICE: resolution selection not implemented yet. defaulting to 0.5°")
    regions = ['glo_30m']

    assert regions == ['glo_30m'], 'invalid region string'
    reg = regions[0]
    t = datetime(kwargs['start'].year, kwargs['start'].month, 1)
    fname = f"multi_1.{reg}.{var}.{t.strftime('%Y%m')}.grb2"
    fetchfile = f"{storage_cfg()}{fname}"

    # if file hasnt been downloaded, fetch it
    if not isfile(fetchfile):  # and kwargs['start'].day == 1:
        logging.info(f'WWIII {kwargs["start"].date().isoformat()} {var}: '
                     f'downloading {fname} from NOAA WaveWatch III...')
        if reg == 'glo_30m' and not 'wind' in var and t.year >= 2018:
            fetchurl = f"{wwiii_src}{t.strftime('%Y/%m')}/gribs/{fname}"
        else:
            fetchurl = f"{wwiii_src}{t.strftime('%Y/%m')}/{reg}/{fname}"
        with requests.get(fetchurl, stream=True, timeout=300) as payload:
            assert payload.status_code == 200, f'couldn\'t retrieve file at {fetchurl}'
            with open(fetchfile, 'wb') as f:
                shutil.copyfileobj(payload.raw, f)

    # open the file, parse data, insert values
    grib = pygrib.open(fetchfile)
    assert grib.messages > 0, f'problem opening {fetchfile}'
    null = 0
    agg = np.array([[], [], [], [], []])
    grbvar = grib[1]['name']
    table = f'{var}{grbvar[0]}' if var == 'wind' else var

    initdb()
    conn, db = database_cfg()
    n1 = db.execute(f"SELECT COUNT(*) FROM {table}").fetchall()[0][0]

    for msg, num in zip(grib, range(1, grib.messages)):
        if (msg['name'] != grbvar and md5(agg.tobytes()).hexdigest() != md5(
                np.array([[], [], [], [], []]).tobytes()).hexdigest()):
            insert(table, agg)
            table = f'{var}{msg["name"][0]}' if var == 'wind' else var
            agg = np.array([[], [], [], [], []])
            grbvar = msg['name']
            null = 0
        if msg.validDate < kwargs['start']:
            continue
        if msg.validDate >= kwargs['end']:
            continue
        z, y, x = msg.data()
        src = np.array(['wwiii' for each in z[~z.mask].data])
        grid = np.vstack(
            (z[~z.mask].data, y[~z.mask], ((x[~z.mask] + 180) % 360) - 180,
             dt_2_epoch([msg.validDate
                         for each in z[~z.mask].data]), src)).astype(object)
        agg = np.hstack((agg, grid))
        null += sum(sum(z.mask))
    insert(table, agg)

    n2 = db.execute(f"SELECT COUNT(*) FROM {table}").fetchall()[0][0]
    conn.close()

    logmsg('wwiii', var, (n1, n2), **kwargs)

    return True


def load_wwiii(var, kwargs):
    """ return downloaded wwiii data for specified wavevar according to given
        time, lat, lon boundaries

        args:
            var: string
                the variable short name of desired wave parameter according to
                WWIII docs the complete list of variable short names can be
                found here (under 'model output')
                https://polar.ncep.noaa.gov/waves/implementations.php
            south, north: float
                ymin, ymax coordinate boundaries (latitude). range: -90, 90
            west, east: float
                xmin, xmax coordinate boundaries (longitude). range: -180, 180
            start: datetime
                the start of the desired time range
            end: datetime
                the end of the desired time range

        return:
            val, lat, lon, epoch as np arrays of floats
    """
    if 'time' in kwargs.keys() and 'start' not in kwargs.keys():
        kwargs['start'] = kwargs['time']
        del kwargs['time']
    if 'end' not in kwargs.keys():
        kwargs['end'] = kwargs['start'] + timedelta(hours=3)

    assert 6 == sum(
        map(lambda kw: kw in kwargs.keys(),
            ['south', 'north', 'west', 'east', 'start', 'end'
             ])), 'malformed query'
    assert var in ('hs', 'tp', 'dp', 'windU', 'windV'), 'invalid load var'

    with index(dx=180,
               dy=90,
               dz=5000,
               dt=timedelta(days=1),
               storagedir=storage_cfg(),
               **{
                   k: v
                   for k, v in kwargs.items()
                   if k in ['south', 'north', 'west', 'east', 'start', 'end']
               }) as fetchmap:
        fetchvar = var if var not in ('windU', 'windV') else 'wind'
        fetchmap(callback=fetch_wwiii, var=fetchvar)

    conn, db = database_cfg()
    db.execute(
        ' AND '.join([
            f'SELECT * FROM {var} WHERE lat >= ?', 'lat <= ?', 'lon >= ?',
            'lon <= ?', 'time >= ?', 'time <= ? '
        ]) + ' ORDER BY time, lat, lon ASC',
        tuple(
            map(str, [
                kwargs['south'], kwargs['north'], kwargs['west'],
                kwargs['east'],
                dt_2_epoch(kwargs['start']),
                dt_2_epoch(kwargs['end'])
            ])))

    slices = np.array(db.fetchall(), dtype=object).T
    if len(slices) == 0:
        logmsg_nodata('wwiii', 'wind', **kwargs)
        return np.array([[], [], [], []])

    conn.close()
    val, lat, lon, epoch, source = slices
    return np.array((val, lat, lon, epoch), dtype=float)


class Wwiii():
    """ collection of module functions for fetching and loading """

    def load_wavedirection(self, **kwargs):
        return load_wwiii('dp', kwargs)

    def load_waveperiod(self, **kwargs):
        return load_wwiii('tp', kwargs)

    def load_windwaveheight(self, **kwargs):
        return load_wwiii('hs', kwargs)

    def load_wind_u(self, **kwargs):
        return load_wwiii('windU', kwargs)

    def load_wind_v(self, **kwargs):
        return load_wwiii('windV', kwargs)

    def load_wind_uv(self, **kwargs):
        with index(dx=180,
                   dy=90,
                   dz=5000,
                   dt=timedelta(days=1),
                   storagedir=storage_cfg(),
                   **kwargs) as fetchmap:
            fetchmap(callback=fetch_wwiii, var='wind')

        sql = ' AND '.join(['SELECT windU.val, windU.lat, windU.lon, windU.time, windV.val FROM windU '\
                'INNER JOIN windV '\
                'ON windU.lat == windV.lat',
                            'windU.lon == windV.lon',
                            'windU.time == windV.time '\
                                    'WHERE windU.lat >= ?',
                            'windU.lat <= ?',
                            'windU.lon >= ?',
                            'windU.lon <= ?',
                            'windU.time >= ?',
                            'windU.time <= ?']) + ' ORDER BY windU.time, windU.lat, windU.lon ASC'
        conn, db = database_cfg()
        db.execute(
            sql,
            tuple(
                map(str, [
                    kwargs['south'], kwargs['north'], kwargs['west'],
                    kwargs['east'],
                    dt_2_epoch(kwargs['start']),
                    dt_2_epoch(kwargs['end'])
                ])))
        qry = np.array(db.fetchall()).T
        conn.close()
        if len(qry) == 0:
            logmsg_nodata('wwiii', 'wind', **kwargs)

            return np.array([[], [], [], [], []])
        wind_u, lat, lon, epoch, wind_v = qry
        val = np.sqrt(np.square(wind_u) + np.square(wind_v))
        return np.array((val, lat, lon, epoch)).astype(float)

    def __str__(self):
        info = '\n'.join([
            "WAVEWATCH III: a third generation wave height,",
            "water depth, and current hindcasting modeling framework.",
            "Developed for the community at NOAA/NCEP. About WWIII:",
            "\thttps://github.com/NOAA-EMC/WW3/wiki/About-WW3",
            "Model descriptions:",
            "\thttps://polar.ncep.noaa.gov/waves/implementations.php"
        ])
        args = "(south, north, west, east, start, end)"
        return str_def(self, info, args)



import pytest
import kadlu
import numpy as np
from datetime import datetime, timedelta
from kadlu.geospatial.data_sources.data_util import reshape_4D
from kadlu.geospatial.data_sources.era5 import clear_cache_era5

# notes:
#  * with default settings, kadlu fetches data in lat/lon bins of 2 degrees and temporal bins of 1 day (24 hours)
#  * ERA5 HRES atmospheric data has a spatial resolution of 0.28125 degrees (31km while wave data have a resolution of 0.36 degrees
#  * ERA4 HRES temporal resolution is 1h

# use @pytest.mark.cds_access as decorater for tests that require access to the CDS API to fetch data


# whether to automatically download data using the CDS API, if the data are not already in the local database
# (should normally be set to False since we don't want to perform API calls every time the tests are run.)
fetch = False


# test region: 1x1 degree area in middle of the north Atlantic; 3-hour time window
kwargs = dict(
    south=36.1,
    west=-39.2,
    north=37.1,
    east=-38.2,
    start=datetime(2024, 7, 1, 8, 0, 0),
    end=datetime(2024, 7, 1, 10, 0, 0),
    top=0,
    bottom=5000,
)


# clear any cached ERA5 data files
clear_cache_era5()


def test_era5_load_wave():
    """ Check that we can load wave data (that have already been fetched) for the test region """
    val, lat, lon, time = kadlu.load(source='era5', var='waveheight', fetch=fetch, **kwargs)
    if len(val) == 0:
        pytest.xfail(
            "Unable to test if ERA5 wave data are loaded correctly "\
            "because data have not been fetched from the Copernicus Climate Data Store."
        )

    assert (len(val) == len(lat) == len(lon))
    assert len(np.unique(time)) == 3 #3 hours
    assert np.all(np.unique(lat) == [36.25, 36.50, 36.75, 37.00])
    assert np.all(np.unique(lon) == [-39.00, -38.75, -38.50, -38.25])
    assert np.all(np.logical_and(val >= 0, val <= 100))


def test_era5_load_irradiance():
    """ Check that we can load irradiance data (that have already been fetched) for the test region """
    val, lat, lon, time = kadlu.load(source='era5', var='irradiance', fetch=fetch, **kwargs)
    if len(val) == 0:
        pytest.xfail(
            "Unable to test if ERA5 solar irradiance data are loaded correctly "\
            "because data have not been fetched from the Copernicus Climate Data Store."
        )

    assert (len(val) == len(lat) == len(lon))
    assert len(np.unique(time)) == 3 #3 hours
    assert np.all(np.unique(time) == [214759.5, 214760.5, 214761.5])
    assert np.all(np.unique(lat) == [36.25, 36.50, 36.75, 37.00])
    assert np.all(np.unique(lon) == [-39.00, -38.75, -38.50, -38.25])
    assert np.all(np.logical_and(val >= 0, val <= 1000))


@pytest.mark.cds_access
def test_era5_fetch_and_load_irradiance():
    """ Check that we can fetch and load irradiance data for the test region """
    val, lat, lon, time = kadlu.load(source='era5', var='irradiance', fetch=True, **kwargs)
    assert (len(val) == len(lat) == len(lon))
    assert len(np.unique(time)) == 3 #3 hours
    assert np.all(np.unique(time) == [214759.5, 214760.5, 214761.5])
    assert np.all(np.unique(lat) == [36.25, 36.50, 36.75, 37.00])
    assert np.all(np.unique(lon) == [-39.00, -38.75, -38.50, -38.25])
    assert np.all(np.logical_and(val >= 0, val <= 1000))


def test_era5_irradiance_is_nonzero_during_daylight():
    """ Check that irradiance data (that have already been fetched) are non-zero during day-light 
        hours for a particular location (Victoria, BC) and date (2024-07-19)
    """
    lat = 48.421
    lon = -123.383
    dt = datetime(2024, 7, 19, 0, 0, 0)
    kwargs = dict(
        south=lat-0.6,
        north=lat+0.6,
        west=lon-0.6,
        east=lon+0.6,
        start=dt,
        end=dt + timedelta(days=1),
    )
    data = kadlu.load(source='era5', var='irradiance', fetch=fetch, **kwargs)
    time = data[3]
    
    if len(time) == 0:
        pytest.xfail(
            "Unable to test if ERA5 solar irradiance data are loaded correctly "\
            "because data have not been fetched from the Copernicus Climate Data Store."
        )

    assert len(np.unique(time)) == 24 #hours
    n = len(np.unique(data[1]))
    m = len(np.unique(data[2]))
    k = len(np.unique(data[3]))
    v = np.reshape(data[0], newshape=(n,m,k), order="F")
    # last axis has UTC hours in increasing order 00, 01, ..., 23
    # Victoria BC is 7 hours behind UTC
    # day-light hours on July 19 are 5:33 AM to 9:05 PM local time
    # so we would expect the following UTC hour bins to have non-zero 
    # solar irradiance values: 0-4, 13-23
    assert np.all(v[0,0,:5] > 0) 
    assert np.all(v[0,0,5:13] == 0.) 
    assert np.all(v[0,0,13:] > 0) 


def test_era5_load_wind():
    """ Check that we can load wind data (that have already been fetched) for the test region """
    val, lat, lon, time = kadlu.load(source='era5', var='wind_uv', fetch=fetch, **kwargs)

    if len(val) == 0:
        pytest.xfail(
            "Unable to test if ERA5 wind data are loaded correctly "\
            "because data have not been fetched from the Copernicus Climate Data Store."
        )

    assert (len(val) == len(lat) == len(lon))
    assert len(np.unique(time)) == 3 #3 hours
    assert np.all(np.unique(lat) == [36.25, 36.50, 36.75, 37.00])
    assert np.all(np.unique(lon) == [-39.00, -38.75, -38.50, -38.25])
    assert np.all(np.logical_and(val >= 0, val <= 100))





""" The ocean module provides an interface to fetching, loading
    and interpolating ocean variables.
"""
import logging
import numpy as np
from scipy.interpolate import NearestNDInterpolator
import kadlu
from kadlu.geospatial.interpolation import get_interpolator, GEOSPATIAL_DIMS
from kadlu.geospatial.data_sources.data_util import fmt_coords, fmt_time
from kadlu.geospatial.data_sources.source_map import load_map, precip_type_map
from kadlu.utils import center_point


# Ocean variable types 
kadlu_vartypes = np.unique([f.rsplit('_', 1)[0] for f in load_map.keys()])


def _drop_dims(data, drop):
    """ Helper function for dropping dimensions and averaging over degenerate data points """
    if drop is None:
        drop = []

    if isinstance(drop, str):
        drop = [drop]

    if len(drop) == 0:
        return data

    v = data.pop("value")
    n_dims = len(data)

    # case I: regular grids
    if np.ndim(v) == 0 or np.ndim(v) == n_dims:
        # remove coordinate arrays for dropped dimensions
        for dim in drop:
            if dim in data:
                del data[dim]

        # average data values over dropped dimensions
        axis = tuple([i for i,d in enumerate(GEOSPATIAL_DIMS) if d in drop and i < np.ndim(v)])
        data["value"] = np.average(v, axis=axis)

    # case II: irregular grids
    else:
        # remove coordinate arrays for dropped dimensions
        for dim in drop:
            if dim in data:
                del data[dim]

        # average data values over dropped dimensions
        # TODO: implement this step
        data["value"] = v

        logger = logging.getLogger("kadlu")
        logger.warning("Averaging over degenerate data points not implemented when dropping dimensions from irregular grids")

    return data


def _validate_loadvars(loadvars):
    """ Check validity of data loading arguments provided to the Ocean class at initialisation.
    
        Args:
            loadvars: dict
                Data loading arguments

        Returns:
            vartypes: list(str)
                Data types

        Raises:
            TypeError: if invalid argments are encountered
    """
    vartypes = []
    for key in loadvars.keys():
        vartype = key.lstrip("load_")
        if vartype not in kadlu_vartypes:
            err_msg = f'{key} is not a valid argument. Valid datasource args include: {", ".join([f"load_{v}" for v in kadlu_vartypes])}'
            raise TypeError(err_msg)
        
        vartypes.append(vartype)

    return vartypes


class Ocean():
    """ Class for retrieving ocean data variables.

        Data will be loaded using the given data sources and 
        geographic, depth, and temporal boundaries.

        It is also possible to write your own data loading function. 
        
        The boundary arguments supplied to the Ocean class will be passed 
        to the the data loading function, i.e., north, south, west, east, 
        top, bottom, start, end.

        TODO: 
            * [ ] Implement averaging across degenerate data points for irregular grids
            * [ ] Re-implement interpolation of precipitation type
            * [ ] Modify data loading classes so they return the data as a dict with keys lat,lon,epoch,depth instead of a numpy array.

        Args:
            north, south: float
                Latitude boundaries, in degrees
            west, east: float
                Longitude boundaries, in degrees
            top, bottom: float
                Depth range, in metres
            start, end: datetime
                UTC time range
            drop: list(str)
                Dimensions to be dropped. If dropping a dimension leads to degeneracy (multiple data 
                points with same coordinates) the average value is used. NOT YET IMPLEMENTED
            interp_args: dict
                Used for passing keyword arguments to the interpolator. See 
                `kadlu.geospatial.interpolation.get_interpolator` for allowed arguments.
            **loadvars:
                Keyword args supplied as 'load_{v}' where v is either an
                integer, float, array of shape [val, lat, lon[, epoch[, depth]]], 
                dict with keys value, lat, lon, epoch, depth, or a string source identifier 
                (e.g. `era5`) as described in the `source_map`

        Attrs:
            origin: tuple(float, float)
                Latitude and longitude coordinates of the centre point of the
                geographic bounding box. This point serves as the origin of the
                planar x-y coordinate system.
            boundaries: dict
                Bounding box for the ocean volume in space and time
            interpolators: dict
                Dictionary of data interpolators
    """
    def __init__(
        self,
        south=kadlu.defaults['south'],
        west=kadlu.defaults['west'],
        north=kadlu.defaults['north'],
        east=kadlu.defaults['east'],
        bottom=kadlu.defaults['bottom'],
        top=kadlu.defaults['top'],
        start=kadlu.defaults['start'],
        end=kadlu.defaults['end'],
        drop=None,
        interp_args=None,
        **loadvars,
    ):
        self.name = self.__class__.__name__

        self.logger = logging.getLogger("kadlu")

        default_value = 0

        if interp_args is None:
            interp_args = dict()
            
        # confirm validity of data loading args
        vartypes = _validate_loadvars(loadvars)

        # ocean spatio-temporal boundaries
        self.boundaries = dict(
            south=south,
            north=north,
            west=west,
            east=east,
            top=top,
            bottom=bottom,
            start=start,
            end=end
        )

        # log info
        info_msg = f"Initializing Ocean in region {fmt_coords(self.boundaries)}"\
                    + f" for time period {fmt_time(self.boundaries)}"\
                    + f" with variables: {vartypes}"
        self.logger.info(info_msg)

        # center point of XY coordinate system
        self.origin = center_point(lat=[south, north], lon=[west, east])

        # load data and initialize interpolators
        self.interpolators = dict()
        for vartype in kadlu_vartypes:

            # get the data loading argument for the given data type;
            # if no argument was provided, use 0 as the default data value
            load_arg = loadvars.get(f"load_{vartype}", default_value)

            # load the data
            if callable(load_arg):
                data = load_arg(**self.boundaries)

            elif isinstance(load_arg, str):
                key = f'{vartype}_{load_arg.lower()}'
                assert key in load_map.keys(), f"No entry found for {key} in Kadlu's load map:\n{load_map}"

                data = load_map[key](**self.boundaries)

            elif isinstance(load_arg, (int, float)):
                data = [load_arg]

            elif isinstance(load_arg, (list, tuple, np.ndarray, dict)):
                if len(load_arg) > 5:
                    err_msg = f'Invalid array shape for load_{vartype}. Arrays must be ordered by [val, lat, lon[, epoch[, depth]]].'
                    raise ValueError(err_msg)

                data = load_arg

            else:
                err_msg = f'Invalid type for load_{vartype}. Valid types include string, float, array, dict, and callable'
                raise TypeError(err_msg)

            # if the data are not already organized into a dict, place the arrays in a dict 
            # using the standard kadlu ordering: value,lat,lon,epoch,depth
            if not isinstance(data, dict):
                keys = ["value"] + GEOSPATIAL_DIMS
                data = {keys[i]: arr for i,arr in enumerate(data)}

            # info message
            v = data["value"]
            if not v is default_value:
                info_msg = f"Finished loading {vartype}"
                for k,arr in data.items():
                    s = arr.shape if np.ndim(arr) > 0 else "scalar"
                    info_msg += f"\n  {k.rjust(5)}: shape={s} min={np.min(arr):.3f} max={np.max(arr):.3f} avg={np.mean(arr):.3f}"

                self.logger.info(info_msg)

            # drop dimensions
            data = _drop_dims(data, drop)

            # pass data to interpolator
            self.interpolators[vartype] = get_interpolator(**data, name=vartype, origin=self.origin, **interp_args)

        """
        # TODO: review this, update as needed
        self.precip_src = loadvars['load_precip_type'] if 'load_precip_type' in loadvars.keys() else None
        """

        # create interpolation method for every variable type
        for vartype,interp in self.interpolators.items():
            setattr(self, vartype, interp)



'''
    def bathymetry_deriv(self, axis, **kwargs):
        """ Interpolates the bathymetric slope along either the south-north (latitude, y) or east-west (longitude, x) axis.

            Args:
                axis: str
                    Axis along which to compute the derivative. Can be `lat`, `lon`, `x`, or `y`

            Returns:
                : array-like
                    The slope values, in m/deg if axis is lat/lon, or dimensionless if axis is x/y
        """
        assert axis in ('lat', 'lon', 'x', 'y'), 'axis must be `lat`, `lon`, `x`, or `y`'

        dlat = (axis == "lat")
        dlon = (axis == "lon")
        dx   = (axis == "x")
        dy   = (axis == "y")

        return self.interps['bathymetry'](dlat=dlat, dlon=dlon, dx=dx, dy=dy, **kwargs)
'''

"""
    def precip_type(self, **kwargs):
        "TODO: provide documentation for this method"
        callback, varmap = precip_type_map[self.precip_src]
        v, y, x, t = callback(west=min(lon),
                              east=max(lon),
                              south=min(lat),
                              north=max(lat),
                              start=self.boundaries['start'],
                              end=self.boundaries['end'])
        return np.array([
            varmap[v]
            for v in NearestNDInterpolator((y, x, t), v)(kwargs["lat"], lon, epoch)
        ])
"""


"""
    API for Era5 dataset from Copernicus Climate Datastore.

    Note: Initial release data (ERA5T) are available about 5 days behind real time.

    Dataset summary:
        https://cds.climate.copernicus.eu/cdsapp#!/dataset/reanalysis-era5-single-levels?tab=overview

    Detailed information about the dataset:
        https://confluence.ecmwf.int/display/CKB/ERA5

    API documentation:
        https://cds.climate.copernicus.eu/toolbox/doc/api.html
"""

import os
import logging
import warnings
import platform
from os.path import isfile, dirname
from glob import glob
from datetime import datetime, timedelta
import cdsapi
import numpy as np

if platform.machine() == 'arm64':
    warnings.warn(
        "ERA5 module requires the pygrib package which is not available for this platform"
    )
else:
    import pygrib

from kadlu import index
from kadlu.geospatial.data_sources.data_util import (
    database_cfg,
    dt_2_epoch,
    logmsg,
    logmsg_nodata,
    storage_cfg,
    str_def,
)

"""
    Names of the tables that will be created in the kadlu geospatial.db database for storing ERA5.

    OBS: Table names must match the variable names used by ERA5 
"""
era5_tables = [
    'significant_height_of_combined_wind_waves_and_swell',
    'mean_wave_direction',
    'mean_wave_period',
    'u_component_of_wind',
    'v_component_of_wind',
    'convective_precipitation',
    'convective_snowfall',
    'normalized_energy_flux_into_ocean',
    'normalized_energy_flux_into_waves',
    'normalized_stress_into_ocean',
    'precipitation_type',
    'surface_solar_radiation_downwards',  #https://codes.ecmwf.int/grib/param-db/169
]


logging.getLogger('cdsapi').setLevel(logging.WARNING)


def initdb():
    """ Create tables in kadlu's geospatial.db database for storing ERA5 data"""
    conn, db = database_cfg()
    for var in era5_tables:
        db.execute(f'CREATE TABLE IF NOT EXISTS {var}'
                   '( val     REAL    NOT NULL, '
                   '  lat     REAL    NOT NULL, '
                   '  lon     REAL    NOT NULL, '
                   '  time    INT     NOT NULL, '
                   '  source  TEXT    NOT NULL) ')
        db.execute(f'CREATE UNIQUE INDEX IF NOT EXISTS '
                   f'idx_{var} on {var}(time, lon, lat, val, source)')
    conn.close()


def clear_cache_era5():
    """ Removes all files with the filename pattern ERA5_*.grb2 in the Kadlu storage directory"""
    logger = logging.getLogger("kadlu")

    # path to folder where Kadlu stores data
    dir_path = storage_cfg()

    if not os.path.exists(dir_path):
        warn_msg = f"Failed to clear ERA5 cache. Kadlu data storage directory not found at {dir_path}."
        logger.warning(warn_msg)
        return
    
    # find all ERA5 grib files
    paths = glob(os.path.join(dir_path, "ERA5_*.grb2"))    

    if len(paths) == 0:
        info_msg = f"ERA5 cache is empty."
        logger.info(info_msg)
        return

    # get their size and remove them
    bytes = 0
    for path in paths:
        bytes += os.path.getsize(path)
        os.remove(path)

    info_msg = f"Emptied ERA5 cache (deleted {len(paths)} files, {bytes/1E6:.1f} MB)"
    logger.info(info_msg)


def fetch_era5(var, *, west, east, south, north, start, **_):
    """ Fetch global ERA5 data for specified variable, geographic region, and time range.

        Downloads 24-hours of global data on the specified day, and saves these data to 
        a *.grb2 file in the kadlu data storage directory, using the recommended spatial 
        resolution of 0.25 x 0.25 degrees.

        The *.grb2 file can be deleted manually by calling the `clear_cache_era5` function 
        to save disk space, if necessary.

        Only data within the specified geographic boundaries (`west`, `east`, `south`, `north`) 
        are inserted into the kadlu geospatial.db database.

        args:
            var: string
                The variable short name of desired wave parameter according to ERA5 docs. 
                The complete list can be found here (table 7 for wave params):
                https://confluence.ecmwf.int/display/CKB/ERA5+data+documentation#ERA5datadocumentation-Temporalfrequency
            west,east,south,north: float
                Geographic boundaries of the data request
            start: datetime.datetime
                UTC date of the data request. 24-hours of data will be fetched.
                
        return:
            True if new data was fetched, else False
    """
    logger = logging.getLogger("kadlu")

    # attempt connecting to the CDS API
    try:
        client = cdsapi.Client()
    except Exception:
        # if attempt fails, raise an error
        raise KeyError('You must obtain a CDS API access token from https://cds-beta.climate.copernicus.eu/how-to-api to download ERA5 data')

    # form request
    # note: we use the recommended spatial resolution of 0.25 x 0.25 degrees
    # https://confluence.ecmwf.int/display/CKB/ERA5%3A+What+is+the+spatial+reference

    t = datetime(start.year, start.month, start.day)
    times = [datetime(t.year, t.month, t.day, h).strftime('%H:00') for h in range(24)]

    request = {
        'product_type': 'reanalysis',
        'format': 'grib',
        'variable': var,
        'year': t.strftime("%Y"),
        'month': t.strftime("%m"),
        'day': t.strftime("%d"),
        'time': times,
        'grid': [0.25, 0.25], 
        'data_format': 'grib',
    }

    # download the data and save as *.grb2 file
    dataset = 'reanalysis-era5-single-levels'
    target = f'ERA5_reanalysis_{var}_{t.strftime("%Y-%m-%d")}.grb2'
    target = f'{storage_cfg()}{target}'
    if not isfile(target):
        logger.info(f'fetching {target}...')
        client.retrieve(dataset, request, target)

    # load the data from the *.grb2 file and insert it into the database
    assert isfile(target)
    grb = pygrib.open(target)
    data = np.array([[], [], [], [], []])
    table = var[4:] if var[0:4] == '10m_' else var

    # process data 'messages' 
    for msg in grb:
        # timestamp
        dt = msg.validDate

        # for forecasts, 'validDate' represents the start time of the forecast (06:00 UTC or 18:00 UTC)
        # so we must add the appropriate number of hourly steps
        # https://confluence.ecmwf.int/pages/viewpage.action?pageId=85402030
        #  
        # for accumulated quantities we should subtracting 1/2 hour to get the time at the center of the 
        # forecast bin; however, since our database stores times as INTs (hours since 2000-01-01), we 
        # instead subtract 1/2 hour when the data is loaded from the database
        
        step = msg.step
        ###if msg.stepType == "accum":
        ###    step -= 0.5

        dt += timedelta(seconds = step * 3600)

        debug_msg = f"[ERA5] Processing message with timestamp {dt} (step={step}) ..."
        logger.debug(debug_msg)

        # read grib data (value, lat, lon)
        z, y, x = msg.data()
        if np.ma.is_masked(z):
            z2 = z[~z.mask].data
            y2 = y[~z.mask]
            x2 = x[~z.mask]
        else:  # wind data has no mask
            z2 = z.reshape(-1)
            y2 = y.reshape(-1)
            x2 = x.reshape(-1)

        # ERA5 uses longitude values in the range [0;360] referenced to the Greenwich Prime Meridian
        # convert to longitude to [-180;180]
        x3 = ((x2 + 180) % 360) - 180

        # index coordinates, select query range subset
        xix = np.logical_and(x3 >= west, x3 <= east)
        yix = np.logical_and(y2 >= south, y2 <= north)
        idx = np.logical_and(xix, yix)

        # collect data in a list (values, lats, lons, times, source)
        msg_data = [
            z2[idx], y2[idx], x3[idx], dt_2_epoch([dt for i in z2[idx]]), ['era5' for i in z2[idx]]
        ]

        # aggregate data
        data = np.hstack((data, msg_data))

    # perform the insertion into the database
    initdb()
    conn, db = database_cfg()
    n1 = db.execute(f"SELECT COUNT(*) FROM {table}").fetchall()[0][0]
    db.executemany(
        f"INSERT OR IGNORE INTO {table} "
        f"VALUES (?,?,?,CAST(? AS INT),?)", data.T)
    n2 = db.execute(f"SELECT COUNT(*) FROM {table}").fetchall()[0][0]
    db.execute("COMMIT")
    conn.commit()
    conn.close()

    # log message
    kwargs = dict(
        south = south,
        west = west,
        north = north,
        east = east,
        start = t,
        end = t + timedelta(hours=24)     
    )
    logmsg('era5', var, (n1, n2), **kwargs)
    
    return True


def load_era5(var, *, west, east, south, north, start, end, fetch=True, **_):
    """ Load ERA5 data from local geospatial.db database

        Args:
            var: str
                Variable to be fetched
            west,east,south,north: float
                Geographic boundaries of the data request
            start: datetime.datetime
                UTC start time for the data request.
            end: datetime.datetime
                UTC end time for the data request.
            fetch: bool
                If the data have not already been downloaded and inserted into 
                Kadlu's local geospatial database, fetch data from the Copernicus 
                Climate Data Store (CDS) automatically using the CDS API. Default is True.
                
        Returns:
            values:
                values of the fetched var
            lat:
                y grid coordinates
            lon:
                x grid coordinates
            epoch:
                timestamps in epoch hours since jan 1 2000
    """
    if fetch:
        # Check local database for data.
        # Fetch data from CDS API, if missing.
        with index(storagedir=storage_cfg(),
                west=west,
                east=east,
                south=south,
                north=north,
                start=start,
                end=end) as fetchmap:
            fetchmap(callback=fetch_era5, var=var)

    # connect to local database
    conn, db = database_cfg()

    # table name in local database
    table = var[4:] if var[0:4] == '10m_' else var  # table name can't start with int

    # check if the table exists
    rows = db.execute(f"SELECT name FROM sqlite_master WHERE type='table' AND name='{table}'").fetchall()
    table_exists = len(rows) > 0

    if table_exists:
        # query local database for data
        sql_query_list = [f"SELECT * FROM {table} WHERE lat >= ?", 'lat <= ?', 'lon >= ?', 'lon <= ?', 'time >= ?', 'time <= ?']
        sql_query = ' AND '.join(sql_query_list) + ' ORDER BY time, lat, lon ASC'
        sql_values = tuple(
            map(str, [south, north, west, east, dt_2_epoch(start), dt_2_epoch(end)])
        )
        db.execute(sql_query, sql_values)
        rowdata = np.array(db.fetchall(), dtype=object).T

    else:
        rowdata = []

    # close database connection
    conn.close()

    # if no data was found, return empty arrays and log info
    if len(rowdata) == 0:
        logmsg_nodata(
            'era5', var,
            west=west, east=east, south=south, north=north,
            start=start, end=end
        )
        return np.array([[], [], [], []])

    val, lat, lon, epoch, source = rowdata
    return np.array((val, lat, lon, epoch), dtype=float)


class Era5():
    """ Collection of module functions for fetching and loading.
    
        The functions return (values, lat, lon, epoch) numpy arrays with 
        shape (num_points, 4) where epoch is the number of hours since 2000-01-01.
    """

    def load_windwaveswellheight(self, **kwargs):
        return load_era5('significant_height_of_combined_wind_waves_and_swell',
                         **kwargs)

    def load_wavedirection(self, **kwargs):
        return load_era5('mean_wave_direction', **kwargs)

    def load_waveperiod(self, **kwargs):
        return load_era5('mean_wave_period', **kwargs)

    def load_precipitation(self, **kwargs):
        return load_era5('convective_precipitation', **kwargs)

    def load_snowfall(self, **kwargs):
        return load_era5('convective_snowfall', **kwargs)

    def load_flux_ocean(self, **kwargs):
        return load_era5('normalized_energy_flux_into_ocean', **kwargs)

    def load_flux_waves(self, **kwargs):
        return load_era5('normalized_energy_flux_into_waves', **kwargs)

    def load_stress_ocean(self, **kwargs):
        return load_era5('normalized_stress_into_ocean', **kwargs)

    def load_precip_type(self, **kwargs):
        return load_era5('precipitation_type', **kwargs)

    def load_wind_u(self, **kwargs):
        return load_era5('10m_u_component_of_wind', **kwargs)

    def load_wind_v(self, **kwargs):
        return load_era5('10m_v_component_of_wind', **kwargs)
    
    def load_insolation(self, **kwargs):
        data = load_era5('surface_solar_radiation_downwards', **kwargs)
        # for accumulated quantities, we subtract 1/2 hour to get the time at the center of the forecast bin
        data[3] -= 0.5
        return data

    def load_irradiance(self, **kwargs):
        data = self.load_insolation(**kwargs)
        data[0] /= 3600
        return data

    def load_wind_uv(self, fetch=True, **kwargs):
        """ Loads wind speed computed as sqrt(wind_u^2 + wind_v^2)"""
        # Check local database for data.
        # Fetch data from CDS API, if missing.
        if fetch:
            with index(storagedir=storage_cfg(),
                    west=kwargs['west'],
                    east=kwargs['east'],
                    south=kwargs['south'],
                    north=kwargs['north'],
                    start=kwargs['start'],
                    end=kwargs['end']) as fetchmap:
                fetchmap(callback=fetch_era5, var='10m_u_component_of_wind')
                fetchmap(callback=fetch_era5, var='10m_v_component_of_wind')

        # establish connection to the geospatial.db database
        conn, db = database_cfg()
        
        # check if the tabls exist
        query = "SELECT name FROM sqlite_master WHERE type='table' AND name IN ('u_component_of_wind','v_component_of_wind')"
        rows = db.execute(query).fetchall()
        tables_exist = len(rows) >= 2

        if not tables_exist:
            return np.array(([], [], [], [])).astype(float)

        # form SQL query
        sql = ' AND '.join(['SELECT u_component_of_wind.val, u_component_of_wind.lat, u_component_of_wind.lon, u_component_of_wind.time, v_component_of_wind.val FROM u_component_of_wind '\
                'INNER JOIN v_component_of_wind '\
                'ON u_component_of_wind.lat == v_component_of_wind.lat',
                            'u_component_of_wind.lon == v_component_of_wind.lon',
                            'u_component_of_wind.time == v_component_of_wind.time '\
                                    'WHERE u_component_of_wind.lat >= ?',
                            'u_component_of_wind.lat <= ?',
                            'u_component_of_wind.lon >= ?',
                            'u_component_of_wind.lon <= ?',
                            'u_component_of_wind.time >= ?',
                            'u_component_of_wind.time <= ?']) + ' ORDER BY u_component_of_wind.time, u_component_of_wind.lat, u_component_of_wind.lon ASC'
        
        # perform the query
        db.execute(
            sql,
            tuple(
                map(str, [
                    kwargs['south'], kwargs['north'], kwargs['west'],
                    kwargs['east'],
                    dt_2_epoch(kwargs['start']),
                    dt_2_epoch(kwargs['end'])
                ])))

        wind_u, lat, lon, epoch, wind_v = np.array(db.fetchall()).T

        # compute speed
        val = np.sqrt(np.square(wind_u) + np.square(wind_v))

        conn.close()

        return np.array((val, lat, lon, epoch)).astype(float)

    def __str__(self):
        info = '\n'.join([
            "Era5 Global Dataset from Copernicus Climate Datastore.",
            "Combines model data with observations from across",
            "the world into a globally complete and consistent dataset",
            "\thttps://cds.climate.copernicus.eu/cdsapp#!/dataset/reanalysis-era5-single-levels"
        ])
        args = "(south, north, west, east, start, end)"
        return str_def(self, info, args)



import logging
from PIL import Image
from functools import reduce
from xml.etree import ElementTree as ET
import json
from datetime import datetime

import matplotlib
'''
matplotlib.use('tkagg')
'''
import mpl_scatter_density
import matplotlib.pyplot as plt
import netCDF4
import numpy as np

from kadlu.geospatial.data_sources.data_util import (
    dt_2_epoch,
    epoch_2_dt,
    index_arr,
)


def load_raster(filepath, plot=False, cmap=None, **kwargs):
    """ load data from raster file

        args:
            filepath: string
                complete filepath descriptor of netcdf file to be read
            plot: boolean
                if True, a plot will be displayed using the qt5agg backend
            cmap: matplotlib colormap
                the full list of available colormaps can be viewed with:
                print(matplotlib.pyplot.colormaps())
                if None is supplied, pyplot will default to
                matplotlib.pyplot.cm.cividis

        returns:
            values: numpy 2D array
            lats:   numpy 1D array
            lons:   numpy 1D array
    """
    if kwargs == {}:
        kwargs.update(dict(south=-90, west=-180, north=90, east=180))

    # load raster
    Image.MAX_IMAGE_PIXELS = 500000000

    #with Image.open(filepath) as im:
    im = Image.open(filepath)
    if 33922 in im.tag.tagdata.keys():
        # GDAL raster format: ModelTiepointTag
        # http://duff.ess.washington.edu/data/raster/drg/docs/geotiff.txt
        i, j, k, x, y, z = im.tag_v2[33922]
        dx, dy, dz = im.tag_v2[33550]  # ModelPixelScaleTag
        meta = im.tag_v2[42112]  # GdalMetadata
        xml = ET.fromstring(meta)
        params = {tag.attrib['name']: tag.text for tag in xml}
        lon = np.arange(x, x + (dx * im.size[0]), dx)
        lat = np.arange(y, y + (dy * im.size[1]), dy)[::-1] - 90
        if np.sum(lat > 91):
            lat -= 90
        rng_lon = (
            index_arr(kwargs['west'], lon),
            index_arr(kwargs['east'], lon),
        )
        rng_lat = (
            abs(index_arr(kwargs['north'], lat[::-1]) - len(lat)),
            abs(index_arr(kwargs['south'], lat[::-1]) - len(lat)),
        )
        logging.debug(
            f'{xml.tag}\nraster coordinate system: {im.tag_v2[34737]}'
            f'\n{json.dumps(params, indent=2, sort_keys=True)}', )
    elif 34264 in im.tag.tagdata.keys():
        # NASA / jet propulsion labs raster format: ModelTransformationTag
        dx, _, _, x, _, dy, _, y, _, _, dz, z, _, _, _, _ = im.tag_v2[34264]
        lon = np.arange(x, x + (dx * im.size[0]), dx)
        lat = np.arange(y, y + (dy * im.size[1]), dy)
        rng_lon = index_arr(kwargs['west'],
                            lon), index_arr(kwargs['east'], lon)
        rng_lat = index_arr(kwargs['south'],
                            -lat), index_arr(kwargs['north'], -lat)

    else:
        assert False, f'error {filepath}: unknown metadata tag encoding'
    assert not (z or dz), f'error {filepath}: 3D rasters not supported yet'

    aoi = im.crop((rng_lon[0], rng_lat[0], rng_lon[1], rng_lat[1])).load()
    grid = np.ndarray((len(range(*rng_lon)), len(range(*rng_lat))))

    n = reduce(np.multiply, (rng_lon[1] - rng_lon[0], rng_lat[1] - rng_lat[0]))
    if n > 10000000:
        logging.info(f'this could take a few moments ({n} points)...')

    for xi in np.arange(rng_lon[0], rng_lon[1]) - rng_lon[0]:
        for yi in np.arange(rng_lat[0], rng_lat[1]) - rng_lat[0]:
            grid[xi][yi] = aoi[int(xi), int(yi)]

    mask = grid == float(im.tag_v2[42113])
    val = np.ma.MaskedArray(grid, mask=mask)

    im.close()

    x1, y1 = np.meshgrid(lon[rng_lon[0]:rng_lon[1]],
                         lat[rng_lat[0]:rng_lat[1]],
                         indexing='ij')
    # plot the data
    if plot:
        fig = plt.figure()
        if (rng_lon[1] - rng_lon[0]) * (rng_lat[1] - rng_lat[0]) >= 100000:
            ax = fig.add_subplot(1, 1, 1, projection='scatter_density')
            plt.axis('scaled')
            _raster = ax.scatter_density(x1, y1, c=val, cmap=cmap)
            plt.tight_layout()
        else:
            ax = fig.add_subplot(1, 1, 1)
            ax.scatter(x1, y1, c=val, cmap=cmap)
            pcm = ax.scatter(x1, y1, marker='s', c=val * -1, cmap=cmap)
            fig.colorbar(pcm, ax=ax)
        plt.show()

    _v = np.array(
        [val[x][y] for x in range(val.shape[0]) for y in range(val.shape[1])],
        dtype=float)
    _x = np.array(
        [x1[x][y] for x in range(x1.shape[0]) for y in range(x1.shape[1])],
        dtype=float)
    _y = np.array(
        [y1[x][y] for x in range(y1.shape[0]) for y in range(y1.shape[1])],
        dtype=float)

    return np.array([_v, _y, _x], dtype=object)


def load_netcdf(filename,
                var=None,
                gridXY=None,
                plot=False,
                cmap=matplotlib.pyplot.cm.bone.reversed(),
                **kwargs):
    """ read environmental data from netcdf and output to gridded numpy array

        args:
            filename: string
                complete filepath descriptor of netcdf file to be read
            var: string (optional)
                the netcdf attribute to be read as the values.
                by default, a guess will be made based on the file metadata
            gridXY: string
                netcdf filename describing grid values if lon/lat are stored as
                grid index
            plot: boolean
                if True, a plot will be displayed using the qt5agg backend
            cmap: matplotlib colormap
                the full list of available colormaps can be viewed with:
                print(matplotlib.pyplot.colormaps())

        returns:
            values: numpy 2D array
            lats:   numpy 1D array
            lons:   numpy 1D array
    """
    if kwargs == {}:
        kwargs.update(
            dict(south=-90,
                 west=-180,
                 north=90,
                 east=180,
                 start=datetime(1, 1, 1),
                 end=datetime.now(),
                 top=0,
                 bottom=9999))
    ncfile = netCDF4.Dataset(filename)

    varmap = dict(
        #MAPSTA='f',  # ifremer tag: appears to be a land mask
        lat='y',
        latitude='y',
        gridY='y',
        lon='x',
        longitude='x',
        gridX='x',
        time='t',
        epoch='t',
        depth='z',
        elevation='z',
    )

    discard = ['crs', 'MAPSTA']

    axes = dict([(varmap[var][0], var) for var in ncfile.variables.keys()
                 if var in varmap.keys()])
    uvars = [
        _ for _ in ncfile.variables.keys()
        if _ not in varmap.keys() and _ not in discard
    ]
    if len(uvars) > 0 or var:
        axes.update({'v': var or uvars[0]})
    if not var:
        var = 'z' if uvars == [] and var is None else uvars[0]

    assert 'x' in axes.keys(), f'missing x axis: {uvars}'
    assert 'y' in axes.keys(), f'missing y axis: {uvars}'
    assert len(uvars) <= 1, f'more than one unknown variable: {uvars}'
    assert sum(
        key in varmap.keys()
        for key in ncfile.variables.keys()) >= len(axes), 'not all vars match'
    assert len(axes) >= len(
        ncfile.variables.keys()) - 1, f'missing axis from: {uvars}'

    logging.info(f'loading data from {ncfile.getncattr("title")}')

    if not gridXY:
        rng_lon = index_arr(kwargs['west'],
                            ncfile[axes['x']][:].data), index_arr(
                                kwargs['east'], ncfile[axes['x']][:].data)
        rng_lat = index_arr(kwargs['south'],
                            ncfile[axes['y']][:].data), index_arr(
                                kwargs['north'], ncfile[axes['y']][:].data)
    else:
        assert False

    out = {} if 'v' not in axes.keys() else dict(
        val=ncfile[axes['v']][:].data[rng_lat[0]:rng_lat[1],
                                      rng_lon[0]:rng_lon[1]])
    out.update(
        dict(lat=ncfile[axes['y']][rng_lat[0]:rng_lat[1]],
             lon=ncfile[axes['x']][rng_lon[0]:rng_lon[1]]))

    # temporal index range
    if 't' in axes.keys():
        if ncfile.variables[
                axes['t']].units == 'days since 1990-01-01T00:00:00Z':
            t0 = datetime(1990, 1, 1)
            rng_t = (index_arr(dt_2_epoch(kwargs['start'], t0),
                               ncfile[axes['t']] * 24),
                     index_arr(dt_2_epoch(kwargs['end'], t0),
                               ncfile[axes['t']] * 24))
            out['time'] = epoch_2_dt(ncfile[axes['t']][rng_t[0]:rng_t[1]] * 24,
                                     t0)
        if ncfile.variables[
                axes['t']].units == 'seconds since 1970-01-01T00:00:00Z':
            t0 = datetime(1970, 1, 1)
            rng_t = (index_arr(dt_2_epoch(kwargs['start'], t0),
                               ncfile[axes['t']] / 60 / 60),
                     index_arr(dt_2_epoch(kwargs['end'], t0),
                               ncfile[axes['t']] / 60 / 60))
            out['time'] = epoch_2_dt(ncfile[axes['t']][rng_t[0]:rng_t[1]],
                                     t0,
                                     unit='seconds')
        else:
            assert False, 'unknown time unit'

    # vertical index range
    if 'z' in axes.keys() and 'v' in axes.keys():
        assert axes['z'] != 'elevation', 'netcdf indexing error'
        rng_z = (index_arr(kwargs['top'], ncfile[axes['z']]),
                 index_arr(kwargs['bottom'], ncfile[axes['z']]))
        out['depth'] = ncfile[axes['z']][rng_z[0]:rng_z[1]]
    elif 'z' in axes.keys() and 'v' not in axes.keys() and len(
            axes.keys()) == 3:
        # when loading bathymetry, z-axis are the intended first column values
        out = dict(val=ncfile[axes['z']][rng_lat[0]:rng_lat[1],
                                         rng_lon[0]:rng_lon[1]],
                   lat=out['lat'].copy(),
                   lon=out['lon'].copy())
        if axes['z'] == 'elevation':
            out['val'] *= -1
    else:
        assert 'v' in axes.keys(), 'something may have gone wrong here...'

    # plot the data
    if plot and len(out.keys()) == 3:
        x1, y1 = np.meshgrid(out['lon'], out['lat'], indexing='xy')
        fig = plt.figure()
        if (rng_lon[1] - rng_lon[0]) * (rng_lat[1] - rng_lat[0]) >= 100000:
            ax = fig.add_subplot(1, 1, 1, projection='scatter_density')
            plt.axis('scaled')
            pcm = ax.scatter_density(x1, y1, c=out['val'], cmap=cmap)
            plt.tight_layout()
        else:
            ax = fig.add_subplot(1, 1, 1)
            pcm = ax.scatter(x1, y1, marker='s', c=out['val'], cmap=cmap)
        fig.colorbar(pcm, ax=ax)
        plt.title('Depth (metres)')
        plt.show()

    lat_m, lon_m = np.meshgrid(out['lat'], out['lon'], indexing='ij')

    grid = np.vstack(
        (out['val'].flatten().astype(float), lat_m.flatten().astype(float),
         lon_m.flatten().astype(float)))

    return grid


'''

kwargs = dict(south = 43.5841631194817, north = 44.01583688051829, west = -59.33904235799604, east = -58.740957642003956, top = 0, bottom = 10000)
filename = '/storage/GEBCO_2020.nc'
var=None
gridXY=None
plot = True
cmap = matplotlib.pyplot.cm.bone.reversed()

'''
